import numpy as np
from scipy import stats
from scipy.optimize import brentq as root


def regTermTest(full_model, restricted_model, ddf, X_names, var_name):
    """
    Performs a LRT for two weighted regression models- one full and one restricted (which nests inside the larger model).
    This is a limited adaption of the same function in the R 'survey' library.
    """
    # Get chisq
    chisq = restricted_model.result.deviance - full_model.result.deviance

    # Get Misspec
    idx = [f"C({var_name})" in n for n in X_names]
    V = full_model.vcov.loc[idx, idx]
    V0 = (full_model.result.cov_params()/full_model.result.scale).loc[idx, idx]
    misspec = np.linalg.eig(np.matmul(np.linalg.pinv(V0), V))[0]

    # Calculate p
    p = _pFsum(x=chisq, df=np.ones(len(misspec)), a=misspec, ddf=ddf)

    return p


def _pFsum(x, df, a, ddf):
    tr = a.mean()
    tr2 = (a**2).mean()/tr**2
    scale = tr * tr2
    ndf = len(a)/tr2

    rval = stats.f.sf(x/ndf/scale, ndf, ddf)

    # Look for saddlepoint
    a = np.append(a, -x/ddf)
    df = np.append(df, ddf)
    if(any(df > 1)):
        a = np.repeat(a, df.tolist(), axis=0)
    s = _saddle(x=0, lam=a)
    if ~np.isnan(s):
        rval = s

    return rval


def _saddle(x, lam):
    """
    This function is used within the pFsum function
    """
    d = max(lam)
    lam = lam / d
    x = x/d

    def k0(z): return -1 * np.log(1-2*z*lam).sum()/2
    def kprime0(z): return (lam/(1-2*z*lam)).sum()
    def kpprime0(z): return 2*(lam**2/(1-2*z*lam)**2).sum()

    if any(lam < 0):
        lam_min = (1/(2*lam[lam < 0])).max() * 0.99999
    elif x > lam.sum():
        lam_min = -0.01
    else:
        lam_min = -1 * len(lam)/(2*x)

    lam_max = (1/(2*lam[lam > 0])).min() * 0.99999

    hatzeta = root(lambda z: kprime0(z) - x, a=lam_min, b=lam_max)

    sign = 1 if hatzeta > 0 else -1
    w = sign * np.sqrt(2*(hatzeta*x-k0(hatzeta)))
    v = hatzeta * np.sqrt(kpprime0(hatzeta))

    if np.abs(hatzeta) < 1e-04:
        return np.nan
    else:
        return 1-stats.norm.cdf(w + np.log(v/w)/w)



from typing import Optional, List

import numpy as np
import pandas as pd
import patsy
import scipy
import statsmodels.api as sm
import statsmodels.formula.api as smf

from ..modules.survey import SurveyDesignSpec, SurveyModel
from .calculations import regTermTest


class Regression(object):
    """
    """
    def __init__(self,
                 variable: str,
                 variable_kind: str,
                 phenotype: str,
                 phenotype_kind: str,
                 data: pd.DataFrame,
                 covariates: Optional[List] = None,
                 survey_design_spec: Optional[SurveyDesignSpec] = None,
                 cov_method: Optional[str] = 'error'):
        # Process input params
        self.variable = variable
        self.variable_kind = variable_kind
        self.phenotype = phenotype
        self.phenotype_kind = phenotype_kind
        self.data = data
        self.covariates = covariates
        self.cov_method = cov_method

        # Subset the data by dropping incomplete cases
        self.data = self.data.dropna(axis='index', how='any', subset=[self.variable, self.phenotype] + self.covariates)

        # Handle survey design
        self.survey_design = None
        if survey_design_spec is not None:
            # Get a survey design object based on the data
            survey_design, index = survey_design_spec.get_survey_design(self.variable, self.data.index)
            # Subset data based on the weights in the design
            self.data = self.data.loc[index]
            # Save the survey design
            self.survey_design = survey_design

        # Select regression family
        if phenotype_kind == "continuous":
            self.family = sm.families.Gaussian(link=sm.families.links.identity())
        elif phenotype_kind == 'binary':
            self.family = sm.families.Binomial(link=sm.families.links.logit())
        else:
            # TODO
            # Note: DoF might change
            raise NotImplementedError("Only continuous and binary phenotypes are currently supported")

        # Set default result values
        self.converged = False
        self.N = len(self.data)
        self.beta = np.nan
        self.SE = np.nan
        self.var_pvalue = np.nan
        self.LRT_pvalue = np.nan
        self.diff_AIC = np.nan
        self.pvalue = np.nan

    def check_covars(self):
        unique_values = self.data[self.covariates].nunique()
        varying_covars = list(unique_values[unique_values > 1].index.values)
        non_varying_covars = list(unique_values[unique_values <= 1].index.values)

        if len(non_varying_covars) > 0:
            print(f"WARNING: {self.variable} has non-varying covariates(s): {', '.join(non_varying_covars)}")
        return varying_covars

    def run(self, min_n):
        """Run the regression and update self with the results"""
        # Check for a minimum amount of data
        if len(self.data) < min_n:
            print(f"{self.variable} = NULL due to: too few complete obervations ({len(self.data)} < {min_n})")
            return
        self.varying_covariates = self.check_covars()

        # Make formulas
        self.formula_restricted = f"{self.phenotype} ~ "
        self.formula_restricted += " + ".join([f"C({var_name})" if str(self.data.dtypes[var_name]) == 'category' else var_name
                                               for var_name in self.varying_covariates])
        if str(self.data.dtypes[self.variable]) == 'category':
            self.formula = self.formula_restricted + f" + C({self.variable})"
        else:
            self.formula = self.formula_restricted + f" + {self.variable}"

        # Run Regression
        if self.survey_design is None:
            if self.variable_kind == 'continuous':
                self.run_continuous()
            elif self.variable_kind == 'binary':
                self.run_binary()  # Essentially same as continuous, except for the string used to key the results
            elif self.variable_kind == "categorical":
                self.run_categorical()
            else:
                raise ValueError(f"Unknown regression variable type '{self.variable_kind}'")
        else:
            if self.variable_kind == 'continuous':
                self.run_continuous_weighted()
            elif self.variable_kind == 'binary':
                self.run_binary_weighted()  # Same as continuous, at least for now
            elif self.variable_kind == "categorical":
                self.run_categorical_weighted()
            else:
                raise ValueError(f"Unknown regression variable type '{self.variable_kind}'")

    def get_results(self):
        """Return a dictionary of the results"""
        return {
            'variable': self.variable,
            'variable_type': self.variable_kind,
            'converged': self.converged,
            'N': self.N,
            'beta': self.beta,
            'SE': self.SE,
            'var_pvalue': self.var_pvalue,
            'LRT_pvalue': self.LRT_pvalue,
            'diff_AIC': self.diff_AIC,
            'pvalue': self.pvalue
        }

    def run_continuous(self):
        # Regress
        est = smf.glm(self.formula, data=self.data, family=self.family).fit(use_t=True)
        # Check convergence
        if not est.converged:
            return
        else:
            self.converged = True
        # Get results
        self.beta = est.params[self.variable]
        self.SE = est.bse[self.variable]
        self.var_pvalue = est.pvalues[self.variable]
        self.pvalue = self.var_pvalue

    def run_binary(self):
        # Regress
        est = smf.glm(self.formula, data=self.data, family=self.family).fit(use_t=True)
        # Check convergence
        if not est.converged:
            return
        else:
            self.converged = True
        # Get results
        # Categorical-type RVs get a different name in the results, and aren't always at the end (since categorical come before non-categorical)
        rv_keys = [k for k in est.params.keys() if self.variable in k]
        try:
            assert len(rv_keys) == 1
            rv_key = rv_keys[0]
        except AssertionError:
            raise KeyError(f"Error extracting results for '{self.variable}', try renaming the variable")
        self.beta = est.params[rv_key]
        self.SE = est.bse[rv_key]
        self.var_pvalue = est.pvalues[rv_key]
        self.pvalue = self.var_pvalue

    def run_categorical(self):
        # Regress both models
        est_restricted = smf.glm(self.formula_restricted, data=self.data, family=self.family).fit(use_t=True)
        est = smf.glm(self.formula, data=self.data, family=self.family).fit(use_t=True)
        # Check convergence
        if not est.converged & est_restricted.converged:
            return
        else:
            self.converged = True
        # Calculate Results
        lrdf = (est_restricted.df_resid - est.df_resid)
        lrstat = -2*(est_restricted.llf - est.llf)
        lr_pvalue = scipy.stats.chi2.sf(lrstat, lrdf)
        self.LRT_pvalue = lr_pvalue
        self.pvalue = self.LRT_pvalue
        self.diff_AIC = est.aic - est_restricted.aic

    def run_continuous_weighted(self):
        y, X = patsy.dmatrices(self.formula, self.data, return_type='dataframe')
        # Create and fit the model
        model = SurveyModel(design=self.survey_design, model_class=sm.GLM, cov_method=self.cov_method,
                            init_args=dict(family=self.family),
                            fit_args=dict(use_t=True))
        model.fit(y=y, X=X)
        # Check convergence
        if not model.result.converged:
            return
        else:
            self.converged = True
        # Get results
        rv_idx_list = [i for i, n in enumerate(X.columns) if self.variable in n]
        if len(rv_idx_list) != 1:
            raise ValueError(f"Failed to find regression variable column in the results for {self.variable}")
        else:
            rv_idx = rv_idx_list[0]
        self.beta = model.params[rv_idx]
        self.SE = model.stderr[rv_idx]
        tval = np.abs(self.beta / self.SE)  # T statistic is the absolute value of beta / SE
        # Get degrees of freedom
        if model.design.has_clusters or model.design.has_strata:
            dof = self.survey_design.get_dof(X)
        else:
            dof = model.result.df_model
        self.var_pvalue = scipy.stats.t.sf(tval, df=dof)*2  # Two-sided t-test
        self.pvalue = self.var_pvalue

    def run_binary_weighted(self):
        y, X = patsy.dmatrices(self.formula, self.data, return_type='dataframe')
        # Create and fit the model
        model = SurveyModel(design=self.survey_design, model_class=sm.GLM, cov_method=self.cov_method,
                            init_args=dict(family=self.family),
                            fit_args=dict(use_t=True))
        model.fit(y=y, X=X)
        # Check convergence
        if not model.result.converged:
            return
        else:
            self.converged = True
        # Get results
        rv_idx_list = [i for i, n in enumerate(X.columns) if self.variable in n]
        if len(rv_idx_list) != 1:
            raise ValueError(f"Failed to find regression variable column in the results for {self.variable}")
        else:
            rv_idx = rv_idx_list[0]
        self.beta = model.params[rv_idx]
        self.SE = model.stderr[rv_idx]
        tval = np.abs(self.beta / self.SE)  # T statistic is the absolute value of beta / SE
        # Get degrees of freedom
        if model.design.has_clusters or model.design.has_strata:
            dof = self.survey_design.get_dof(X)
        else:
            dof = model.result.df_model
        self.var_pvalue = scipy.stats.t.sf(tval, df=dof)*2  # Two-sided t-test
        self.pvalue = self.var_pvalue

    def run_categorical_weighted(self):
        # The change in deviance between a model and a nested version (with n fewer predictors) follows a chi-square distribution with n DoF
        # See https://en.wikipedia.org/wiki/Deviance_(statistics)
        # Regress restricted model
        y, X_restricted = patsy.dmatrices(self.formula_restricted, self.data, return_type='dataframe')
        model_restricted = SurveyModel(design=self.survey_design, model_class=sm.GLM, cov_method=self.cov_method,
                                       init_args=dict(family=self.family),
                                       fit_args=dict(use_t=True))
        model_restricted.fit(y=y, X=X_restricted)
        # Regress full model (Already have the survey_design and index objects)
        y, X = patsy.dmatrices(self.formula, self.data, return_type='dataframe')
        model = SurveyModel(design=self.survey_design, model_class=sm.GLM, cov_method=self.cov_method,
                            init_args=dict(family=self.family),
                            fit_args=dict(use_t=True))
        model.fit(y=y, X=X)
        # Check convergence
        if not model.result.converged & model_restricted.result.converged:
            return
        else:
            self.converged = True
        # Calculate Results
        dof = self.survey_design.get_dof(X)
        if model.design.has_strata or model.design.has_clusters:
            # Calculate pvalue using vcov
            lr_pvalue = regTermTest(full_model=model, restricted_model=model_restricted, ddf=dof, X_names=X.columns, var_name=self.variable)
        else:
            # Calculate using llf from model results
            lrdf = (model_restricted.result.df_resid - model.result.df_resid)
            lrstat = -2*(model_restricted.result.llf - model.result.llf)
            lr_pvalue = scipy.stats.chi2.sf(lrstat, lrdf)
        # Gather Other Results
        self.LRT_pvalue = lr_pvalue
        self.pvalue = self.LRT_pvalue
        self.diff_AIC = model.result.aic - model_restricted.result.aic



import numpy as np
import pandas as pd
import statsmodels


class SurveyModel(object):
    """

    Parameters
    -------
    design : Instance of class SurveyDesign
    model_class : Instance of class GLM
    cov_method : str
        Method for calculating covariance
    init_args : Dictionary of arguments
        when initializing the model
    fit_args : Dictionary of arguments
        when fitting the model

    Attributes
    ----------
    design : Instance of class SurveyDesign
    model : Instance of class GLM
    init_args : Dictionary of arguments
        when initializing the model
    fit_args : Dictionary of arguments
        when fitting the model
    params : (p, ) array
        Array of coefficients of model
    vcov : (p, p) array
        Covariance matrix
    stderr : (p, ) array
        Standard error of cofficients
    """
    def __init__(self, design, model_class, cov_method='stata', init_args={}, fit_args={}):
        self.design = design
        self.model = model_class
        self.cov_method = cov_method
        self.init_args = dict(init_args)
        self.fit_args = dict(fit_args)

        if self.model is statsmodels.genmod.generalized_linear_model.GLM:
            self.glm_flag = True
        else:
            self.glm_flag = False

    def _stata_linearization_vcov(self, X, y):
        """
        Get the linearized covariance matrix using STATA's methodology

        Parameters
        ----------
        X : array-like
            A n x p array where 'n' is the number of observations and 'p'
            is the number of regressors.
        y : array-like
            1d array of the response variable

        Returns
        -------
        vcov : (p,p) array
            The covariance matrix

        Notes
        -----
        This uses a 'sandwich' method, where three matrices are multiplied
        together. The outer parts of the 'sandwich' is the inverse of the
        negative hessian. The inside is the design-based variance of the
        estimation of a total, ie np.dot(weights, observations) where
        'observations' is the derivative of the log pseudolikelihood w.r.t
        theta (X*Beta) multiplied by the data

        Reference
        ---------
        http://www.stata.com/manuals13/svyvarianceestimation.pdf
        """
        # Input data indicies must match design indicies
        design_index = self.design.strat.index  # May be all ones, but it always exists
        assert X.index.equals(design_index)
        assert y.index.equals(design_index)

        # Get hessian
        hessian = pd.DataFrame(self.initialized_model.hessian(self.params, observed=True),
                               index=self.params.index,
                               columns=self.params.index)
        hess_inv = pd.DataFrame(np.linalg.inv(hessian.values),
                                index=self.params.index,
                                columns=self.params.index)

        # Get the first derivative of the loglikelihood function evaluated at params for each observation
        d_hat = pd.DataFrame(self.initialized_model.score_obs(self.params),
                             index=self.design.strat.index,
                             columns=self.params.index)

        # Add strat and clust information to the d_hat index
        d_hat = pd.merge(d_hat, self.design.strat, left_index=True, right_index=True).set_index('strat', append=True)
        d_hat = pd.merge(d_hat, self.design.clust, left_index=True, right_index=True).set_index('clust', append=True)

        # Get sum of d_hat within each cluster
        jdata = d_hat.groupby(axis=0, level='clust').apply(sum)

        # Add strata label to jdata
        jdata = pd.merge(jdata, self.design.strat_for_clust[jdata.index].rename('strat'), left_index=True, right_index=True)\
                  .set_index('strat', append=True)

        def center_strata(data, single_cluster, pop_mean):
            if len(data) > 1 or single_cluster == 'scaled' or single_cluster == 'certainty':
                # Substract the mean across clusters
                # This results in a value of 0 for the strata when there is only 1 cluster
                return data - data.mean()
            elif len(data) == 1 and single_cluster == 'centered':
                # Subtract the population mean from the single-cluster strata
                return data - pop_mean
            elif len(data) == 1:
                raise ValueError(f"Strat '{data.name}' has a single cluster. "
                                 f"Adjust the 'single_cluster' SurveyDesign parameter or reassign the cluster to avoid this error.")

        # Center each stratum
        single_cluster = self.design.single_cluster  # 'scaled', 'certainty', or 'centered'.  Anything else will throw an error for single-cluster-strata
        jdata = jdata.groupby(axis=0, level='strat').apply(lambda g: center_strata(g, single_cluster, d_hat.mean()))

        # Scale after centering, if required
        if single_cluster == 'scaled':
            single_cluster_scale = self.design.n_strat / (self.design.n_strat - sum(self.design.clust_per_strat == 1))
            jdata *= single_cluster_scale

        nh = self.design.clust_per_strat.loc[self.design.strat_for_clust].astype(np.float64)
        mh = np.sqrt(nh / (nh-1))
        mh[mh == np.inf] = 1  # Replace infinity with 1.0 (due to single cluster where nh==1)
        fh = np.sqrt(1 - self.design.fpc)
        jdata *= (fh[:, None] * mh[:, None])
        v_hat = np.dot(jdata.T, jdata)
        vcov = np.dot(hess_inv, v_hat).dot(hess_inv.T)
        # Return as a data.frame
        vcov = pd.DataFrame(vcov, index=self.params.index, columns=self.params.index)
        return vcov

    def _jackknife_vcov(self, X, y):
        # Calculate parameter values, leaving out one cluster at a time
        replicate_params = dict()
        for c in self.design.clust_names:
            w = self.design.get_jackknife_rep_weights(dropped_clust=c)
            init_args = {k: v for k, v in self.init_args.items()}
            if self.glm_flag:
                init_args["freq_weights"] = w
            else:
                init_args["weights"] = w
            replicate_params[c] = self._calc_params(y, X, init_args, self.fit_args)

        # Collect data into a dataframe and center it
        replicate_params = pd.DataFrame.from_dict(replicate_params, orient='index', columns=self.params.index)
        replicate_params.index.name = 'clust'
        self.replicate_params = replicate_params - self.params

        nh = self.design.clust_per_strat.loc[self.design.strat_for_clust].astype(np.float64)
        mh = np.sqrt((nh - 1) / nh)
        mh[mh == np.inf] = 1  # Replace infinity with 1.0 (due to single cluster where nh==1)
        fh = np.sqrt(1 - self.design.fpc)
        self.replicate_params *= (mh[:, None] * fh[:, None])

        vcov = np.dot(self.replicate_params.T, self.replicate_params)

        # Return as a data.frame
        vcov = pd.DataFrame(vcov, index=self.params.index, columns=self.params.index)
        return vcov

    def fit(self, y, X, center_by='est'):

        assert y.index.equals(X.index)
        self.center_by = center_by

        # Get weights (will be an array of ones if not provided)
        weights = self.design.weights
        # Normalize weights
        weights = weights/weights.mean()
        # Use weights in the regression
        if self.glm_flag:
            self.init_args["freq_weights"] = weights
        else:
            self.init_args["weights"] = weights
        self.params = self._get_params(y, X)

        if self.design.has_strata or self.design.has_clusters:
            # Calculate stderr based on covariance
            if self.cov_method == 'jackknife':
                self.vcov = self._jackknife_vcov(X, y)
            elif self.cov_method == 'stata':
                self.vcov = self._stata_linearization_vcov(X, y)
            else:
                return ValueError(f"cov_method '{self.cov_method}' is not supported")
            if self.vcov.ndim == 2:
                self.stderr = np.sqrt(np.diag(self.vcov))
            else:
                self.stderr = np.sqrt(self.vcov)
        else:
            # Use the stderr calculated in the original weighted regression
            self.stderr = self.result.bse

    def _get_params(self, y, X):
        self.initialized_model = self.model(y, X, **self.init_args)
        self.result = self.initialized_model.fit(**self.fit_args)
        return self.result.params

    def _calc_params(self, y, X, init_args, fit_args):
        """Like _get_params, but doesn't modify self.  Jackknife calculates params many times, but must keep original values."""
        initialized_model = self.model(y, X, **init_args)
        result = initialized_model.fit(**fit_args)
        return result.params



import click
import pandas as pd

from ...modules.survey import SurveyDesignSpec
from ...modules import analyze, io
from ..parameters import input_file, output_file


@click.group(name='analyze')
def analyze_cli():
    pass


@analyze_cli.command(help="Run an EWAS analysis")
@click.argument('phenotype', type=click.STRING)
@click.argument('bin-data', type=input_file)
@click.argument('cat-data', type=input_file)
@click.argument('cont-data', type=input_file)
@click.option('--covariate', '-c', multiple=True, help="Covariates")
@click.option('--covariance-calc', default='stata', type=click.Choice(['stata', 'jackknife']), help="Covariance calculation method")
@click.option('--min-n', default=200, type=click.IntRange(0, 999999), help="Minimum number of complete cases needed to run a regression")
@click.option('--survey-data', type=input_file, default=None,
              help="Tab-separated data file with survey weights, strata IDs, and/or cluster IDs.  Must have an 'ID' column.")
@click.option('--strata', type=click.STRING, default=None, help="Name of the strata column in the survey data")
@click.option('--cluster', type=click.STRING, default=None, help="Name of the cluster column in the survey data")
@click.option('--nested/--not-nested', help="Whether survey data is nested or not")
@click.option('--weights-file', type=input_file, default=None,
              help="Tab-delimited data file with 'Variable' and 'weight' columns to match weights from the survey data to specific variables")
@click.option('--weight', '-w', type=click.STRING, default=None,
              help="Name of a survey weight column found in the survey data.  This option can't be used with --weights-file")
@click.option('--single-cluster', type=click.Choice(['error', 'scaled', 'centered', 'certainty']), default='error', help="How to handle singular clusters")
@click.argument('output', type=output_file)
def ewas(phenotype, bin_data, cat_data, cont_data, covariate, covariance_calc, min_n,
         survey_data, strata, cluster, nested, weights_file, weight, single_cluster, output):
    """Run EWAS and add corrected pvalues"""
    # Load data
    bin_data = io.load_data(bin_data)
    cat_data = io.load_data(cat_data)
    cont_data = io.load_data(cont_data)
    # Make covariates into a list
    covariates = list(covariate)
    # Load optional survey data
    if survey_data is not None:
        survey_data = io.load_data(survey_data)
        if weights_file is not None and weight is not None:
            raise ValueError("Either 'weights-file' or 'weight' should be specified, not both.")
        elif weights_file is not None:
            weights = pd.read_csv(weights_file, sep="\t")
            if list(weights) != ['variable', 'weight']:
                raise ValueError(f"The weights-file must be a tab-separated file with two columns: 'variable' and 'weight'. "
                                 f"Columns were: {', '.join(list(weights))}")
            weights = weights.set_index('variable')['weight'].to_dict()
        elif weight is not None:
            weights = weight
        elif weights_file is None and weight is None:
            weights = None
        sd = SurveyDesignSpec(survey_data, strata=strata, cluster=cluster, nest=nested, weights=weights, single_cluster=single_cluster)
    else:
        sd = None
    # Remove variables with missing weights
    if type(weights) == dict:
        missing_weights_bin = set(list(bin_data)) - set([phenotype] + covariates) - set(weights.keys())
        for v in missing_weights_bin:
            click.echo(click.style(f"\tSkipping binary variable '{v}' because it wasn't listed in the weights file", fg='yellow'))
            bin_data = bin_data.drop(v, axis='columns')
        missing_weights_cat = set(list(cat_data)) - set([phenotype] + covariates) - set(weights.keys())
        for v in missing_weights_cat:
            click.echo(click.style(f"\tSkipping categorical variable '{v}' because it wasn't listed in the weights file", fg='yellow'))
            cat_data = cat_data.drop(v, axis='columns')
        missing_weights_cont = set(list(cont_data)) - set([phenotype] + covariates) - set(weights.keys())
        for v in missing_weights_cont:
            click.echo(click.style(f"\tSkipping continuous variable '{v}' because it wasn't listed in the weights file", fg='yellow'))
            cont_data = cont_data.drop(v, axis='columns')
    # Run ewas
    result = analyze.ewas(phenotype=phenotype, covariates=covariates, bin_df=bin_data, cat_df=cat_data, cont_df=cont_data,
                          survey_design_spec=sd, cov_method=covariance_calc, min_n=min_n)
    # Add corrected pvalues
    analyze.add_corrected_pvalues(result)
    # Save
    result.to_csv(output, sep="\t")
    # Log
    click.echo(click.style(f"Done: Saved EWAS results to {output}", fg='green'))


@analyze_cli.command(help="filter out non-significant results")
@click.argument('ewas_result_data', type=input_file)
@click.argument('output', type=output_file)
@click.option('--fdr/--bonferroni', 'use_fdr', default=True, help="Use FDR (--fdr) or Bonferroni pvalues (--bonferroni).  FDR by default.")
@click.option('--pvalue', '-p', type=click.FLOAT, default=0.05, help="Keep results with a pvalue <= this value (0.05 by default)")
def get_significant(ewas_result_data, output, use_fdr, pvalue):
    # Load data
    data = pd.read_csv(ewas_result_data, sep="\t", index_col=['variable', 'phenotype'])
    # Check columns
    if list(data) != analyze.result_columns + analyze.corrected_pvalue_columns:
        raise ValueError(f"{ewas_result_data} was not a valid EWAS result file.")
    # Filter
    if use_fdr:
        col = 'pvalue_fdr'
    else:
        col = 'pvalue_bonferroni'
    data = data.loc[data[col] <= pvalue, ]
    # Save result
    io.save(data, filename=output)
    # Log
    click.echo(click.style(f"Done: Saved {len(data):,} variables to {output}", fg='green'))



import click

input_file = click.Path(exists=True, file_okay=True, dir_okay=False, readable=True)
output_file = click.Path(file_okay=True, dir_okay=False, writable=True)


def process_skip_only(ctx, param, value):
    if len(value) == 0:
        return None
    else:
        return list(value)


skip = click.option('-s', '--skip', type=click.STRING, multiple=True, callback=process_skip_only,
                    help="variables to skip")
only = click.option('-o', '--only', type=click.STRING, multiple=True, callback=process_skip_only,
                    help="variables to process, skipping all others")



import pytest
import statsmodels as sm

# Datasets for testing
@pytest.fixture
def plantTraits():
    data = sm.datasets.get_rdataset('plantTraits', 'cluster', cache=True).data
    data.index.name = "ID"
    return data



"""
IO
========

Input/Output of data in different formats

  .. autosummary::
     :toctree: modules/io

     load_data
     save
     load_dtypes
     save_dtypes
"""

from pathlib import Path
import json
from typing import Optional, Union

import pandas as pd

from ..internal.utilities import get_dtypes, set_dtypes


def load_data(filename: str, index_col: Optional[Union[str, int]] = 0, sep: str = '\t',
              dtypes: Optional[Union[str, bool]] = None, **kwargs):
    """Load data from a file

    Wraps the Pandas 'read_csv' function but requires an index_col,
    and reports the number of variables and observations that were loaded.

    Parameters
    ----------
    filename: str or Path
        File with data to be used in CLARITE
    index_col: int or string (default 0)
        Column to use as the row labels of the DataFrame.
    sep: str (default is "\t" for tab-separated)
        column separator (delimiter)
    dtypes: optional string or boolean
        Name of the dtypes file for this data.
         If not provided, check for the default dtypes file (filename + ".dtypes") and use it if available.
         If a boolean, require that the default dtypes file is available (True) or don't use it (False).
    **kwargs:
        Other keword arguments to pass to pd.read_csv

    Returns
    -------
    DataFrame
        The index column will be used when merging

    Examples
    --------
    Load a tab-delimited file with an "ID" column

    >>> df = io.load_data('nhanes.txt', index_col="ID", sep="\t")
    Loaded 22,624 observations of 970 variables
    """
    # Load data
    data = pd.read_csv(filename, index_col=index_col, sep=sep, **kwargs)
    print(f"Loaded {len(data):,} observations of {len(data.columns):,} variables")

    # Rename index to ID
    data.index.name = "ID"

    # Get default dtypes filename if one wasn't provided
    if dtypes is None or dtypes is True or dtypes is False:
        dtypes_filename = str(filename) + ".dtypes"
    else:
        dtypes_filename = dtypes
    dtypes_file = Path(dtypes_filename)

    # Depending on the input parameter, handle dtypes
    if dtypes is False:
        # Don't load the file no matter what
        print("\tIgnoring any existing dtypes file, using default datatypes instead")
    elif dtypes is None:
        # Try to load the default filename, but do nothing if not found
        if dtypes_file.exists():
            dtypes = load_dtypes(dtypes_filename)
            set_dtypes(data, dtypes)
            print(f"\tLoaded dtypes file: {dtypes_filename}")
    elif dtypes is True or type(dtypes) == str:
        # Throw an error if the file isn't found (True, or a specific file)
        if dtypes_file.exists():
            dtypes = load_dtypes(dtypes_filename)
            set_dtypes(data, dtypes)
        else:
            raise ValueError(f"Couldn't load dtypes file: {dtypes_filename}")

    return data


def load_dtypes(filename: str):
    """
    Load  a dtypes file

    Parameters
    ----------
    filename: str
        Name of the file to be loaded

    Returns
    -------
    Series
        A Series of datatypes

    Examples
    --------
    >>> clarite.io.load_dtypes('data/test_data.dtypes')
    """
    fpath = Path(filename)
    if not fpath.exists():
        raise ValueError(f"Could not read '{filename}'")
    else:
        with fpath.open('r') as f:
            try:
                dtypes = json.load(f)
            except json.JSONDecodeError as e:
                raise ValueError(f"'{filename}' was not a valid dtypes file: {e}")
    return dtypes


def save_dtypes(data: pd.DataFrame, filename: str):
    """
    Save a datatype file (.dtype) for the given data

    Parameters
    ----------
    data: pd.DataFrame
        Data to be saved
    filename: str
        Name of data file to be used in CLARITE - the 'dtypes' extension is added automatically if needed.

    Returns
    -------
    None

    Examples
    --------
    >>> clarite.io.save_dtypes(df, 'data/test_data')
    """
    # Check filename
    if not filename.endswith(".dtypes"):
        filename += ".dtypes"

    # Save
    dtypes = get_dtypes(data)
    with open(filename, 'w') as f:
        json.dump(dtypes, f)


def save(data: pd.DataFrame, filename: str):
    """
    Save a data to a file along with a dtype file

    Parameters
    ----------
    data: pd.DataFrame
        Data to be saved
    filename: str
        File with data to be used in CLARITE

    Returns
    -------
    None

    Examples
    --------
    >>> clarite.io.save(df, 'data/test_data')
    """
    # Check filename
    if not filename.endswith(".txt"):
        filename += ".txt"
    filename_dtypes = filename + ".dtypes"

    if len(data.columns) > 0:
        data.to_csv(filename, sep="\t", header=True)
        save_dtypes(data, filename_dtypes)



import click

from .commands import analyze_cli, describe_cli, io_cli, modify_cli, plot_cli, process_cli


@click.group()
def entry_point():
    pass


entry_point.add_command(analyze_cli)
entry_point.add_command(describe_cli)
entry_point.add_command(io_cli)
entry_point.add_command(modify_cli)
entry_point.add_command(plot_cli)
entry_point.add_command(process_cli)



from pathlib import Path
import click
import pandas as pd
from ...modules import process, io
from ..parameters import input_file, output_file, skip, only


@click.group(name='process')
def process_cli():
    pass


@process_cli.command(help="Categorize data based on the number of unique values")
@click.argument('data', type=input_file)
@click.argument('output', type=output_file)
@click.option('--cat_min', default=3, help="Minimum number of unique values in a variable to make it a categorical type")
@click.option('--cat_max', default=6, help="Maximum number of unique values in a variable to make it a categorical type")
@click.option('--cont_min', default=15, help="Minimum number of unique values in a variable to make it a continuous type")
def categorize(data, output, cat_min, cat_max, cont_min):
    # Get output prefix
    if output is None:
        output = Path(data)
    else:
        output = Path(output)
    # Load data
    data = io.load_data(data)
    # Categorize
    df_bin, df_cat, df_cont, df_check = process.categorize(data, cat_min=cat_min, cat_max=cat_max, cont_min=cont_min)
    # Save Data
    for name, df in zip(['binary', 'categorical', 'continuous', 'check'], [df_bin, df_cat, df_cont, df_check]):
        output_name = str(output.with_suffix('')) + f"_{name}"
        if len(df.columns) > 0:
            io.save(df, filename=output_name)
            click.echo(click.style(f"Saved {name} results to {output_name}.txt", fg='green'))
        else:
            click.echo(click.style(f"No {name} variables available to save to {output_name}", fg='yellow'))


@process_cli.command(help="Merge variables from two different datasets into one")
@click.argument('left', type=input_file)
@click.argument('right', type=input_file)
@click.argument('output', type=output_file)
@click.option('--how', '-h', default='outer', type=click.Choice(['left', 'right', 'inner', 'outer']), help="Type of Merge")
def merge_variables(left, right, output, how):
    # Load data
    left = io.load_data(left)
    right = io.load_data(right)
    # Merge
    result = process.merge_variables(left, right, how)
    # Save
    io.save(result, filename=output)
    # Log
    click.echo(click.style(f"Done: Saved {len(result.columns):,} with {len(result):,} variables to {output}", fg='green'))


@process_cli.command(help="Merge rows from two different datasets into one")
@click.argument('top', type=input_file)
@click.argument('bottom', type=input_file)
@click.argument('output', type=output_file)
def merge_rows(top, bottom, output):
    # Load data
    top = io.load_data(top)
    bottom = io.load_data(bottom)
    # Merge
    extra = set(list(top)) - set(list(bottom))
    missing = set(list(bottom)) - set(list(top))
    if len(extra) > 0:
        raise ValueError(f"Couldn't merge rows: Extra columns in the 'bottom' data: {', '.join(extra)}")
    elif len(missing) > 0:
        raise ValueError(f"Couldn't merge rows: Missing columns in the 'bottom' data: {', '.join(missing)}")
    elif (top.dtypes != bottom.dtypes).any():
        raise ValueError("Couldn't merge rows: different data types")
    else:
        result = pd.concat([top, bottom], verify_integrity=True, sort=False)
    # Save
    io.save(result, filename=output)
    # Log
    click.echo(click.style(f"Done: Saved {len(result.columns):,} with {len(result):,} variables to {output}", fg='green'))


@process_cli.command(help="Move variables from one dataset to another")
@click.argument('left', type=input_file)
@click.argument('right', type=input_file)
@click.argument('output_left', type=output_file)
@click.argument('output_right', type=output_file)
@skip
@only
def move_variables(left, right, output_left, output_right, skip, only):
    # Load data
    left = io.load_data(left)
    right = io.load_data(right)
    before = len(list(left))
    # Move
    left, right = process.move_variables(left, right, skip=skip, only=only)
    after = len(list(left))
    # Save
    io.save(left, filename=output_left)
    io.save(right, filename=output_right)
    # Log
    click.echo(click.style(f"Done: Moved {before-after:,} variables and saved results to {output_left} and {output_right}", fg='green'))



"""
Plot
========

Functions that generate plots

  .. autosummary::
     :toctree: modules/plot

     histogram
     distributions
     manhattan
"""


from copy import copy
import datetime
from typing import Dict, List, Optional, Tuple

from matplotlib.collections import PatchCollection
import matplotlib.pyplot as plt
from matplotlib.patches import Rectangle
from matplotlib.backends.backend_pdf import PdfPages
import seaborn as sns
from statsmodels.api import qqplot
import numpy as np
import pandas as pd

from .analyze import result_columns, corrected_pvalue_columns
from .._version import get_versions

clarite_version = get_versions()


def histogram(
    data,
    column: str,
    figsize: Tuple[int, int] = (12, 5),
    title: Optional[str] = None,
    **kwargs,
):
    """
    Plot a histogram of the values in the given column.  Takes kwargs for seaborn's distplot.

    Parameters
    ----------
    data: pd.DataFrame
        The DataFrame containing data to be plotted
    column: string
        The name of the column that will be plotted
    figsize: tuple(int, int), default (12, 5)
        The figure size of the resulting plot
    title: string or None, default None
        The title used for the plot
    **kwargs:
        Other keyword arguments to pass to the distplot function of Seaborn

    Returns
    -------
    None

    Examples
    --------
    >>> import clarite
    >>> title = f"Discovery: Skew of BMIMBX = {stats.skew(nhanes_discovery_cont['BMXBMI']):.6}"
    >>> clarite.plot.histogram(nhanes_discovery_cont, column="BMXBMI", title=title, bins=100)

    .. image:: ../../_static/plot/histogram.png
    """
    if title is None:
        title = f"Histogram for {column}"
    if column not in data.columns:
        raise ValueError("'column' must be an existing column in the DataFrame")

    _, ax = plt.subplots(figsize=figsize)
    ax.set_title(title)
    sns.distplot(data.loc[~data[column].isna(), column], ax=ax, **kwargs)


def distributions(
    data,
    filename: str,
    continuous_kind: str = "count",
    nrows: int = 4,
    ncols: int = 3,
    quality: str = "medium",
    variables: Optional[List[str]] = None,
    sort: bool = True,
):
    """
    Create a pdf containing histograms for each binary or categorical variable, and one of several types of plots for each continuous variable.

    Parameters
    ----------
    data: pd.DataFrame
        The DataFrame containing data to be plotted
    filename: string
        Name of the saved pdf file.  The extension will be added automatically if it was not included.
    continuous_kind: string
        What kind of plots to use for continuous data.  Binary and Categorical variables will always be shown with histograms.
        One of {'count', 'box', 'violin', 'qq'}
    nrows: int (default=4)
        Number of rows per page
    ncols: int (default=3)
        Number of columns per page
    quality: 'low', 'medium', or 'high'
        Adjusts the DPI of the plots (150, 300, or 1200)
    variables: List[str] or None
        Which variables to plot.  If None, all variables are plotted.
    sort: Boolean (default=True)
        Whether or not to sort variable names

    Returns
    -------
    None

    Examples
    --------
    >>> import clarite
    >>> clarite.plot.distributions(df[['female', 'occupation', 'LBX074']], filename="test")

    .. image:: ../../_static/plot/distributions_count.png

    >>> clarite.plot.distributions(df[['female', 'occupation', 'LBX074']], filename="test", continuous_kind='box')

    .. image:: ../../_static/plot/distributions_box.png

    >>> clarite.plot.distributions(df[['female', 'occupation', 'LBX074']], filename="test", continuous_kind='violin')

    .. image:: ../../_static/plot/distributions_violin.png

    >>> clarite.plot.distributions(df[['female', 'occupation', 'LBX074']], filename="test", continuous_kind='qq')

    .. image:: ../../_static/plot/distributions_qq.png

    """
    # Limit variables
    if variables is not None:
        data = data[variables]

    # Check filename
    if not filename.endswith(".pdf"):
        filename += ".pdf"

    # Set DPI
    dpi_dict = {"low": 150, "medium": 300, "high": 1200}
    dpi = dpi_dict.get(quality, None)
    if dpi is None:
        raise ValueError(f"quality was set to '{quality}' which is not a valid value")

    # Make sure file is writeable
    try:
        with PdfPages(filename) as pdf:
            pass
    except OSError:
        raise OSError(f"Unable to write to '{filename}'")

    with PdfPages(filename) as pdf:
        # Determine the number of pages
        plots_per_page = nrows * ncols
        total_pages = (len(data.columns) + (plots_per_page - 1)) // plots_per_page
        print(f"Generating a {total_pages} page PDF for {len(data.columns):,} variables")
        # Starting plot space
        page_num = 1
        row_idx = 0
        col_idx = 0
        # Loop through all variables
        if sort:
            variables = sorted(list(data))
        else:
            variables = list(data)
        for variable in variables:
            if row_idx == 0 and col_idx == 0:
                # New Page
                _ = plt.subplots(squeeze=False, figsize=(8.5, 11), dpi=dpi)
                plt.suptitle(f"Page {page_num}")
            # Plot non-NA values and record the number of those separately (otherwise they can cause issues with generating a KDE)
            ax = plt.subplot2grid((nrows, ncols), (row_idx, col_idx))
            if str(data.dtypes[variable]) == "category":
                sns.countplot(data.loc[~data[variable].isna(), variable], ax=ax)
            else:
                if continuous_kind == "count":
                    sns.distplot(
                        data.loc[~data[variable].isna(), variable],
                        kde=False,
                        norm_hist=False,
                        hist_kws={"alpha": 1},
                        ax=ax,
                    )
                elif continuous_kind == "box":
                    sns.boxplot(data.loc[~data[variable].isna(), variable], ax=ax)
                elif continuous_kind == "violin":
                    sns.violinplot(data.loc[~data[variable].isna(), variable], ax=ax)
                elif continuous_kind == "qq":
                    # QQ plots have to be sub-sampled otherwise there are too many points and the pdf is blank
                    d = data.loc[~data[variable].isna(), variable]
                    if len(d) > 400:
                        d = d.sample(n=400, random_state=1)
                    qqplot(d, line="s", fit=True, ax=ax, color="steelblue", alpha=0.7)
                else:
                    raise ValueError(
                        "Unknown value for 'continuous_kind': must be one of {'count', 'box', 'violin', 'qq'}"
                    )
            # Update xlabel with NA information
            na_count = data[variable].isna().sum()
            ax.set_xlabel(
                f"{variable}\n{na_count:,} of {len(data[variable]):,} are NA ({na_count/len(data[variable]):.2%})"
            )
            # Move to next plot space
            col_idx += 1
            if col_idx == ncols:  # Wrap to next row
                col_idx = 0
                row_idx += 1
            if row_idx == nrows:  # Wrap to next page
                row_idx = 0
                page_num += 1
                # Save the current page
                plt.tight_layout(rect=[0, 0.03, 1, 0.95])
                pdf.savefig()
                plt.close()
        # Save final page, unless a full page was finished and the page_num is now more than total_pages
        if page_num == total_pages:
            plt.tight_layout(rect=[0, 0.03, 1, 0.95])
            pdf.savefig()
            plt.close()
        # Add metadata
        d = pdf.infodict()
        d["Title"] = "Multipage PDF Example"
        d["Author"] = f"CLARITE {clarite_version}"
        d["Subject"] = "Distribution plots"
        d["CreationDate"] = datetime.datetime.today()
        d["ModDate"] = datetime.datetime.today()


def manhattan(
    dfs: Dict[str, pd.DataFrame],
    categories: Dict[str, str] = dict(),
    num_labeled: int = 3,
    label_vars: List[str] = list(),
    figsize: Tuple[int, int] = (12, 6),
    dpi: int = 300,
    title: Optional[str] = None,
    colors: List[str] = ["#53868B", "#4D4D4D"],
    background_colors: List[str] = ["#EBEBEB", "#FFFFFF"],
    filename: Optional[str] = None,
):
    """
    Create a Manhattan-like plot for a list of EWAS Results

    Parameters
    ----------
    dfs: DataFrame
        Dictionary of dataset names to pandas dataframes of ewas results (requires certain columns)
    categories: dictionary (string: string)
        A dictionary mapping each variable name to a category name
    num_labeled: int, default 3
        Label the top <num_labeled> results with the variable name
    label_vars: list of strings, default empty list
        Label the named variables
    figsize: tuple(int, int), default (12, 6)
        The figure size of the resulting plot in inches
    dpi: int, default 300
        The figure dots-per-inch
    title: string or None, default None
        The title used for the plot
    colors: List(string, string), default ["#53868B", "#4D4D4D"]
        A list of colors to use for alternating categories (must be same length as 'background_colors')
    background_colors: List(string, string), default ["#EBEBEB", "#FFFFFF"]
        A list of background colors to use for alternating categories (must be same length as 'colors')
    filename: Optional str
        If provided, a copy of the plot will be saved to the specified file

    Returns
    -------
    None

    Examples
    --------
    >>> clarite.plot_manhattan({'discovery':disc_df, 'replication':repl_df}, categories=data_categories, title="EWAS Results")

    .. image:: ../../_static/plot/manhattan.png
    """
    # Hardcoded options
    offset = 5  # Spacing between categories

    # Parameter Validation - EWAS Results
    for df_idx, df in enumerate(dfs.values()):
        if list(df) != result_columns + corrected_pvalue_columns:
            raise ValueError(
                f"This plot may only be created for EWAS results with corrected p-values added. "
                f"DataFrame {df_idx + 1} of {len(dfs)} did not have the expected columns."
            )

    # Create a dataframe of pvalues indexed by variable name
    df = (
        pd.DataFrame.from_dict({k: v["pvalue"] for k, v in dfs.items()})
        .stack()
        .reset_index()
    )
    df.columns = ("variable", "phenotype", "dataset", "pvalue")
    df[["variable", "phenotype", "dataset"]] = df[
        ["variable", "phenotype", "dataset"]
    ].astype("category")

    # Add category
    df["category"] = (
        df["variable"].apply(lambda v: categories.get(v, "unknown")).astype("category")
    )

    # Transform pvalues
    df["-log10(p value)"] = -1 * df["pvalue"].apply(np.log10)

    # Sort and extract an 'index' column
    df = df.sort_values(["category", "variable"]).reset_index(drop=True)

    # Update index (actually x position) column by padding between each category
    df["category_x_offset"] = (
        df.groupby("category").ngroup() * offset
    )  # sorted category number, multiplied to pad between categories
    df["xpos"] = (
        df.groupby(["category", "variable"]).ngroup() + df["category_x_offset"]
    )  # sorted category/variable number plus category offset

    # Create Plot and structures to hold category info
    fig, axes = plt.subplots(
        len(dfs), 1, figsize=figsize, dpi=dpi, sharex=True, sharey=True
    )
    x_labels = []
    x_labels_pos = []
    foreground_rectangles = [list() for c in colors]

    # Wrap axes in a list if there is only one plot so that it can be iterated
    if len(dfs) == 1:
        axes = [axes]

    # Plot
    if len(categories) > 0:
        # Include category info
        for category_num, (category_name, category_data) in enumerate(
            df.groupby("category")
        ):
            # background bars
            left = category_data["xpos"].min() - (offset / 2) - 0.5
            right = category_data["xpos"].max() + (offset / 2) + 0.5
            width = right - left
            for ax in axes:
                ax.axvspan(
                    left,
                    right,
                    facecolor=background_colors[category_num % len(colors)],
                    alpha=1,
                )
            # foreground bars
            rect = Rectangle(xy=(left, -1), width=width, height=1)
            foreground_rectangles[category_num % len(colors)].append(rect)
            # plotted points
            for dataset_num, (dataset_name, dataset_data) in enumerate(
                category_data.groupby("dataset")
            ):
                if (
                    len(dataset_data) > 0
                ):  # Sometimes a category has no variables in one dataset
                    dataset_data.plot(
                        kind="scatter",
                        x="xpos",
                        y="-log10(p value)",
                        label=None,
                        color=colors[category_num % len(colors)],
                        zorder=2,
                        s=10000 / len(df),
                        ax=axes[dataset_num],
                    )
            # Record centered position and name of xtick for the category
            x_labels.append(category_name)
            x_labels_pos.append(left + width / 2)

        # Show the foreground rectangles for each category by making a collection then adding it to each axes
        for rect_idx, rect_list in enumerate(foreground_rectangles):
            pc = PatchCollection(
                rect_list,
                facecolor=colors[rect_idx],
                edgecolor=colors[rect_idx],
                alpha=1,
                zorder=2,
            )
            for ax in axes:
                ax.add_collection(
                    copy(pc)
                )  # have to add a copy since the same object can't be added to multiple axes
    else:
        # Just plot variables
        for dataset_num, (dataset_name, dataset_data) in enumerate(
            df.groupby("dataset")
        ):
            # Plot points
            dataset_data.plot(
                kind="scatter",
                x="xpos",
                y="-log10(p value)",
                label=None,
                color=colors[0],
                zorder=2,
                s=10000 / len(df),
                ax=axes[dataset_num],
            )
        # Plot a single rectangle on each axis
        for ax in axes:
            rect = Rectangle(xy=(0, -1), width=df["xpos"].max(), height=1)
            pc = PatchCollection(
                [rect], facecolor=colors[0], edgecolor=colors[0], alpha=1, zorder=2
            )
            ax.add_collection(pc)

    # Format plot
    for dataset_num, (dataset_name, dataset_data) in enumerate(df.groupby("dataset")):
        ax = axes[dataset_num]
        # Set title
        ax.set_title(dataset_name, fontsize=14)
        ax.yaxis.label.set_size(16)
        # Update y-axis
        plt.yticks(fontsize=8)
        # Label points
        top_n = dataset_data.sort_values("-log10(p value)", ascending=False).head(
            n=num_labeled
        )["variable"]
        labeled = set(label_vars) | set(top_n)
        for _, row in dataset_data.loc[
            dataset_data["variable"].isin(labeled),
        ].iterrows():
            ax.text(
                row["xpos"] + 1,
                row["-log10(p value)"],
                str(row["variable"]),
                rotation=0,
                ha="left",
                va="center",
            )

    # Format bottom axes
    ax = axes[-1]
    ax.set_xticks(x_labels_pos)
    ax.set_xticklabels(x_labels)
    ax.tick_params(labelrotation=90)
    ax.set_xlim([df["xpos"].min() - offset, df["xpos"].max() + offset])
    ax.set_ylim([-1, df["-log10(p value)"].max() + 10])
    ax.set_xlabel("")  # Hide x-axis label since it is obvious

    # Significance line for each dataset
    for dataset_num, (dataset_name, dataset_data) in enumerate(df.groupby("dataset")):
        num_tests = dataset_data["pvalue"].count()
        significance = -np.log10(0.05 / num_tests)
        axes[dataset_num].axhline(
            y=significance,
            color="red",
            linestyle="-",
            zorder=3,
            label=f"0.05 Bonferroni with {num_tests} tests",
        )
        axes[dataset_num].legend(
            loc="upper right", bbox_to_anchor=(1, 1.1), fancybox=True, shadow=True
        )

    # Title
    fig.suptitle(title, fontsize=20)

    # Save
    if filename is not None:
        plt.savefig(filename, bbox_inches="tight")
    plt.show()



"""
Modify
======

Functions used to filter and/or change some data, always taking in one set of data and returning one set of data.

  .. autosummary::
     :toctree: modules/modify

     colfilter_percent_zero
     colfilter_min_n
     colfilter_min_cat_n
     rowfilter_incomplete_obs
     recode_values
     remove_outliers
     make_binary
     make_categorical
     make_continuous

"""

from typing import Optional, List, Union

import numpy as np
import pandas as pd

from ..internal.utilities import _validate_skip_only


def colfilter_percent_zero(data: pd.DataFrame, filter_percent: float = 90.0,
                           skip: Optional[Union[str, List[str]]] = None, only: Optional[Union[str, List[str]]] = None):
    """
    Remove columns which have <proportion> or more values of zero (excluding NA)

    Parameters
    ----------
    data: pd.DataFrame
        The DataFrame to be processed and returned
    filter_percent: float, default 90.0
            If the percentage of rows in the data with a value of zero is greater than or equal to this value, the variable is filtered out.
    skip: str, list or None (default is None)
        List of variables that the filter should *not* be applied to
    only: str, list or None (default is None)
        List of variables that the filter should *only* be applied to

    Returns
    -------
    data: pd.DataFrame
        The filtered DataFrame

    Examples
    --------
    >>> import clarite
    >>> nhanes_discovery_cont = clarite.modify.colfilter_percent_zero(nhanes_discovery_cont)
    Removed 30 of 369 variables (8.13%) which were equal to zero in at least 90.00% of non-NA observations.
    """
    columns = _validate_skip_only(list(data), skip, only)
    num_before = len(data.columns)

    percent_value = 100 * data.apply(lambda col: sum(col == 0) / col.count())
    kept = (percent_value < filter_percent) | ~data.columns.isin(columns)
    num_removed = num_before - sum(kept)

    print(f"Removed {num_removed:,} of {num_before:,} variables ({num_removed/num_before:.2%}) "
          f"which were equal to zero in at least {filter_percent:.2f}% of non-NA observations.")
    return data.loc[:, kept]


def colfilter_min_n(data: pd.DataFrame, n: int = 200,
                    skip: Optional[Union[str, List[str]]] = None, only: Optional[Union[str, List[str]]] = None):
    """
    Remove columns which have less than <n> unique values (excluding NA)

    Parameters
    ----------
    data: pd.DataFrame
        The DataFrame to be processed and returned
    n: int, default 200
        The minimum number of unique values required in order for a variable not to be filtered
    skip: str, list or None (default is None)
        List of variables that the filter should *not* be applied to
    only: str, list or None (default is None)
        List of variables that the filter should *only* be applied to

    Returns
    -------
    data: pd.DataFrame
        The filtered DataFrame

    Examples
    --------
    >>> import clarite
    >>> nhanes_discovery_bin = clarite.modify.colfilter_min_n(nhanes_discovery_bin)
    Removed 129 of 361 variables (35.73%) which had less than 200 values
    """
    columns = _validate_skip_only(list(data), skip, only)
    num_before = len(data.columns)

    counts = data.count()  # by default axis=0 (rows) so counts number of non-NA rows in each column
    kept = (counts >= n) | ~data.columns.isin(columns)
    num_removed = num_before - sum(kept)

    print(f"Removed {num_removed:,} of {num_before:,} variables ({num_removed/num_before:.2%}) which had less than {n} values")
    return data.loc[:, kept]


def colfilter_min_cat_n(data, n: int = 200, skip: Optional[Union[str, List[str]]] = None, only: Optional[Union[str, List[str]]] = None):
    """
    Remove columns which have less than <n> occurences of each unique value

    Parameters
    ----------
    data: pd.DataFrame
        The DataFrame to be processed and returned
    n: int, default 200
        The minimum number of occurences of each unique value required in order for a variable not to be filtered
    skip: str, list or None (default is None)
        List of variables that the filter should *not* be applied to
    only: str, list or None (default is None)
        List of variables that the filter should *only* be applied to

    Returns
    -------
    data: pd.DataFrame
        The filtered DataFrame

    Examples
    --------
    >>> import clarite
    >>> nhanes_discovery_bin = clarite.modify.colfilter_min_cat_n(nhanes_discovery_bin)
    Removed 159 of 232 variables (68.53%) which had a category with less than 200 values
    """
    columns = _validate_skip_only(list(data), skip, only)
    num_before = len(data.columns)

    min_category_counts = data.apply(lambda col: col.value_counts().min())
    kept = (min_category_counts >= n) | ~data.columns.isin(columns)
    num_removed = num_before - sum(kept)

    print(f"Removed {num_removed:,} of {num_before:,} variables ({num_removed/num_before:.2%}) which had a category with less than {n} values")
    return data.loc[:, kept]


def rowfilter_incomplete_obs(data, skip: Optional[Union[str, List[str]]] = None, only: Optional[Union[str, List[str]]] = None):
    """
    Remove rows containing null values

    Parameters
    ----------
    data: pd.DataFrame
        The DataFrame to be processed and returned
    skip: str, list or None (default is None)
        List of columns that are not checked for null values
    only: str, list or None (default is None)
        List of columns that are the only ones to be checked for null values

    Returns
    -------
    data: pd.DataFrame
        The filtered DataFrame

    Examples
    --------
    >>> import clarite
    >>> nhanes = clarite.modify.rowfilter_incomplete_obs(only=[phenotype] + covariates)
    Removed 3,687 of 22,624 rows (16.30%) due to NA values in the specified columns
    """
    columns = _validate_skip_only(list(data), skip, only)

    keep_IDs = data[columns].isnull().sum(axis=1) == 0  # Number of NA in each row is 0
    n_removed = len(data) - sum(keep_IDs)

    print(f"Removed {n_removed:,} of {len(data):,} rows ({n_removed/len(data):.2%}) due to NA values in the specified columns")
    return data[keep_IDs]


def recode_values(data, replacement_dict,
                  skip: Optional[Union[str, List[str]]] = None, only: Optional[Union[str, List[str]]] = None):
    """
    Convert values in a dataframe.  By default, replacement occurs in all columns but this may be modified with 'skip' or 'only'.
    Pandas has more powerful 'replace' methods for more complicated scenarios.

    Parameters
    ----------
    data: pd.DataFrame
        The DataFrame to be processed and returned
    replacement_dict: dictionary
        A dictionary mapping the value being replaced to the value being inserted
    skip: str, list or None (default is None)
        List of variables that the replacement should *not* be applied to
    only: str, list or None (default is None)
        List of variables that the replacement should *only* be applied to

    Examples
    --------
    >>> import clarite
    >>> clarite.recode_values(df, {7: np.nan, 9: np.nan}, only=['SMQ077', 'DBD100'])
    Replaced 17 values from 22,624 rows in 2 columns
    >>> clarite.recode_values(df, {10: 12}, only=['SMQ077', 'DBD100'])
    No occurences of replaceable values were found, so nothing was replaced.
    """
    # Limit columns if needed
    if skip is not None or only is not None:
        columns = _validate_skip_only(list(data), skip, only)
        replacement_dict = {c: replacement_dict for c in columns}

    # Replace
    result = data.replace(to_replace=replacement_dict, value=None, inplace=False)

    # Log
    diff = result.eq(data)
    diff[pd.isnull(result) & pd.isnull(data)] = True  # NAs are not equal by default
    diff = ~diff  # make True where a value was replaced
    cols_with_changes = (diff.sum() > 0).sum()
    cells_with_changes = diff.sum().sum()
    if cells_with_changes > 0:
        print(f"\tReplaced {cells_with_changes:,} values from {len(data):,} rows in {cols_with_changes:,} columns")
    else:
        print(f"\tNo occurences of replaceable values were found, so nothing was replaced.")

    # Return
    return result


def remove_outliers(data, method: str = 'gaussian', cutoff=3,
                    skip: Optional[Union[str, List[str]]] = None, only: Optional[Union[str, List[str]]] = None):
    """
    Remove outliers from the dataframe by replacing them with np.nan

    Parameters
    ----------
    data: pd.DataFrame
        The DataFrame to be processed and returned
    method: string, 'gaussian' (default) or 'iqr'
        Define outliers using a gaussian approach (standard deviations from the mean) or inter-quartile range
    cutoff: positive numeric, default of 3
        Either the number of standard deviations from the mean (method='gaussian') or the multiple of the IQR (method='iqr')
        Any values equal to or more extreme will be replaced with np.nan
    skip: str, list or None (default is None)
        List of variables that the replacement should *not* be applied to
    only: str, list or None (default is None)
        List of variables that the replacement should *only* be applied to

    Examples
    --------
    >>> import clarite
    >>> nhanes_rm_outliers = clarite.remove_outliers(nhanes, method='iqr', cutoff=1.5, only=['DR1TVB1', 'URXP07', 'SMQ077'])
    Removing outliers with values < 1st Quartile - (1.5 * IQR) or > 3rd quartile + (1.5 * IQR) in 3 columns
        430 of 22,624 rows of URXP07 were outliers
        730 of 22,624 rows of DR1TVB1 were outliers
        Skipped filtering 'SMQ077' because it is a categorical variable
    >>> nhanes_rm_outliers = clarite.remove_outliers(only=['DR1TVB1', 'URXP07'])
    Removing outliers with values more than 3 standard deviations from the mean in 2 columns
        42 of 22,624 rows of URXP07 were outliers
        301 of 22,624 rows of DR1TVB1 were outliers
    """
    # Copy to avoid replacing in-place
    data = data.copy(deep=True)

    # Which columns
    columns = _validate_skip_only(list(data), skip, only)

    # Check cutoff and method, printing what is being done
    if cutoff <= 0:
        raise ValueError("'cutoff' must be >= 0")
    if method == 'iqr':
        print(f"Removing outliers with values < 1st Quartile - ({cutoff} * IQR) or > 3rd quartile + ({cutoff} * IQR) in {len(columns):,} columns")
    elif method == 'gaussian':
        print(f"Removing outliers with values more than {cutoff} standard deviations from the mean in {len(columns):,} columns")
    else:
        raise ValueError(f"'{method}' is not a supported method for outlier removal - only 'gaussian' and 'iqr'.")

    # Process each column
    for c in columns:
        if str(data.dtypes[c]) == 'category':
            print(f"\tSkipped filtering '{c}' because it is a categorical variable")
            continue
        # Calculate outliers
        if method == 'iqr':
            q1 = data[c].quantile(0.25)
            q3 = data[c].quantile(0.75)
            iqr = abs(q3 - q1)
            bottom = q1 - (iqr * cutoff)
            top = q3 + (iqr * cutoff)
            outliers = (data[c] < bottom) | (data[c] > top)
        elif method == 'gaussian':
            outliers = (data[c]-data[c].mean()).abs() > cutoff*data[c].std()
        # If there are any, log a message and replace them with np.nan
        if sum(outliers) > 0:
            print(f"\t{sum(outliers):,} of {len(data):,} rows of {c} were outliers")
            data.loc[outliers, c] = np.nan

    return data


def make_binary(data: pd.DataFrame, skip: Optional[Union[str, List[str]]] = None, only: Optional[Union[str, List[str]]] = None):
    """
    Validate and type a dataframe of binary variables

    Checks that each variable has at most 2 values and converts the type to pd.Categorical

    Parameters
    ----------
    data: pd.DataFrame or pd.Series
        Data to be processed
    skip: str, list or None (default is None)
        List of variables that should *not* be made binary
    only: str, list or None (default is None)
        List of variables that are the *only* ones to be made binary

    Returns
    -------
    data: pd.DataFrame
        DataFrame with the same data but validated and converted to binary types

    Examples
    --------
    >>> import clarite
    >>> df = clarite.modify.make_binary(df)
    Set 32 of 32 variables as binary, each with 4,321 observations
    """
    # Which columns
    columns = _validate_skip_only(list(data), skip, only)

    # Check the number of unique values
    unique_values = data.nunique()
    num_non_binary = (unique_values[columns] > 2).sum()
    if num_non_binary > 0:
        raise ValueError(f"{num_non_binary} of {len(columns)} variables did not have 2 unique values and couldn't be processed as a binary type")
    # TODO: possibly add further validation to make sure values are 1 and 0

    # Convert dtype
    data = data.astype({c: 'category' for c in columns})
    print(f"Set {len(columns):,} of {len(data.columns)} variables as binary, each with {len(data):,} observations")

    return data


def make_categorical(data: pd.DataFrame, skip: Optional[Union[str, List[str]]] = None, only: Optional[Union[str, List[str]]] = None):
    """
    Validate and type a dataframe of categorical variables

    Converts the type to pd.Categorical

    Parameters
    ----------
    data: pd.DataFrame or pd.Series
        Data to be processed
    skip: str, list or None (default is None)
        List of variables that should *not* be made categorical
    only: str, list or None (default is None)
        List of variables that are the *only* ones to be made categorical

    Returns
    -------
    data: pd.DataFrame
        DataFrame with the same data but validated and converted to categorical types

    Examples
    --------
    >>> import clarite
    >>> df = clarite.modify.make_categorical(df)
    Set 12 of 12 variables as categorical, each with 4,321 observations
    """
    # Which columns
    columns = _validate_skip_only(list(data), skip, only)

    # TODO: possibly add further validation

    # Convert dtype
    data = data.astype({c: 'category' for c in columns})
    print(f"Set {len(columns):,} of {len(data.columns)} variables as categorical, each with {len(data):,} observations")

    return data


def make_continuous(data: pd.DataFrame, skip: Optional[Union[str, List[str]]] = None, only: Optional[Union[str, List[str]]] = None):
    """
    Validate and type a dataframe of continuous variables

    Converts the type to numeric

    Parameters
    ----------
    data: pd.DataFrame or pd.Series
        Data to be processed
    skip: str, list or None (default is None)
        List of variables that should *not* be made continuous
    only: str, list or None (default is None)
        List of variables that are the *only* ones to be made continuous

    Returns
    -------
    data: pd.DataFrame
        DataFrame with the same data but validated and converted to numeric types

    Examples
    --------
    >>> import clarite
    >>> df = clarite.modify.make_continuous(df)
    Set 128 of 128 variables as continuous, each with 4,321 observations
    """
    # Which columns
    columns = _validate_skip_only(list(data), skip, only)

    # TODO: possibly add further validation

    # Convert dtype
    data = pd.DataFrame({c: data[c] if c not in columns else pd.to_numeric(data[c]) for c in list(data)})
    print(f"Set {len(columns):,} of {len(data.columns)} variables as continuous, each with {len(data):,} observations")

    return data



from typing import Optional, List, Union

import numpy as np
import pandas as pd


def _validate_skip_only(columns, skip: Optional[Union[str, List[str]]] = None, only: Optional[Union[str, List[str]]] = None):
    """Validate use of the 'skip' and 'only' parameters, returning a valid list of columns to filter"""
    # Convert string to a list
    if type(skip) == str:
        skip = [skip]
    if type(only) == str:
        only = [only]

    if skip is not None and only is not None:
        raise ValueError(f"It isn't possible to specify 'skip' ({skip}) and 'only' ({only}) at the same time.")
    elif skip is not None and only is None:
        invalid_cols = set(skip) - set(columns)
        if len(invalid_cols) > 0:
            raise ValueError(f"Invalid columns passed to 'skip': {', '.join(invalid_cols)}")
        columns = [c for c in columns if c not in set(skip)]
    elif skip is None and only is not None:
        invalid_cols = set(only) - set(columns)
        if len(invalid_cols) > 0:
            raise ValueError(f"Invalid columns passed to 'only': {', '.join(invalid_cols)}")
        columns = [c for c in columns if c in set(only)]

    if len(columns) == 0:
        raise ValueError("No columns available for filtering")

    return columns


def get_dtypes(data):
    """
    Convert dtypes of a DataFrame or Series to a dictionary format
    Examples:
      binary: {'female': {'type': 'category', 'categories': [0, 1], 'ordered': False}}
      categorical: {'CALCIUM_Unknown': {'type': 'category', 'categories': [0.0, 0.066666666, 0.933333333], 'ordered': False}}
      continuous: {'BMXBMI': {'type': 'float64'}}
    """
    dtypes = {variable_name: {'type': str(dtype)} if str(dtype) != 'category'
              else {'type': str(dtype), 'categories': list(dtype.categories.values.tolist()), 'ordered': dtype.ordered}
              for variable_name, dtype in data.dtypes.iteritems()}
    return dtypes


def set_dtypes(data, dtypes):
    """
    Set the dtypes of a dataframe according to a dtypes dictionary (in-place)
    """
    # Validate
    missing_types = set(list(data)) - set(dtypes.keys())
    extra_dtypes = set(dtypes.keys()) - set(list(data))
    if len(missing_types) > 0:
        raise ValueError(f"Dtypes file is missing some values: {', '.join(missing_types)}")
    if len(extra_dtypes) > 0:
        raise ValueError(f"Dtypes file has types for variables not found in the data: {', '.join(extra_dtypes)}")

    for col in list(data):
        typeinfo = dtypes[col]
        newtype = typeinfo['type']
        if typeinfo['type'] == 'category':
            newtype = pd.CategoricalDtype(categories=np.array(typeinfo['categories']), ordered=typeinfo['ordered'])
        data[col] = data[col].astype(newtype)



import click
from ...modules import io
from ...internal.utilities import _validate_skip_only
from ..parameters import input_file, output_file, skip, only


@click.group(name='io')
def io_cli():
    pass


@io_cli.command(help="Load data from a tab-separated or comma-separated file and save it in the standard format")
@click.argument('data', type=input_file)
@click.argument('output', type=output_file)
@click.option('--index', '-i', type=click.STRING, help="Name of the column to use as the index.  Default is the first column.")
@click.option('--tab', is_flag=True, help="File is tab-separated")
@click.option('--comma', is_flag=True, help="File is comma-separated")
@skip
@only
def load_data(data, output, index, tab, comma, skip, only):
    """Load Data from some format and save it in the standard format for further processing"""
    # Determine seperator character
    if tab and comma:
        raise ValueError(f"Can't specify '--tab' and '--comma' at the same time")
    elif not tab and not comma:
        raise ValueError(f"Must specify either '--tab' or '--comma'")
    elif tab and not comma:
        sep = "\t"
    elif comma and not tab:
        sep = ","
    # Index column is assumed to be the first column if not provided
    if index is None:
        index = 0
    # Load Data, ignoring any existing dtypes file
    data = io.load_data(filename=data, index_col=index, sep=sep, dtypes=False)
    # Raise error if no variables were found (only one column)
    if len(data.columns) == 0:
        raise ValueError("Only one column was found- was the correct delimeter used?")
    # Process skip/only parameters
    columns = _validate_skip_only(list(data), skip, only)
    data = data[columns]
    # Save
    io.save(data, filename=output)
    # Log
    click.echo(click.style(f"Done: Saved {len(data.columns):,} variables with {len(data):,} observations to {output}", fg='green'))



"""
Process
========

Functions used to process data from one form into another, such as categorizing variables and placing them in separate DataFrames

  .. autosummary::
     :toctree: modules/process

     categorize
     merge_variables
     move_variables

"""

from typing import List, Optional, Union

import pandas as pd

from ..internal.utilities import _validate_skip_only


def categorize(data: pd.DataFrame, cat_min: int = 3, cat_max: int = 6, cont_min: int = 15):
    """
    Divide variables into binary, categorical, continuous, and ambiguous dataframes

    Parameters
    ----------
    data: pd.DataFrame
        The DataFrame to be processed
    cat_min: int, default 3
        Minimum number of unique, non-NA values for a categorical variable
    cat_max: int, default 6
        Maximum number of unique, non-NA values for a categorical variable
    cont_min: int, default 15
        Minimum number of unique, non-NA values for a continuous variable

    Returns
    -------
    bin_df: pd.DataFrame
        DataFrame with variables that were categorized as *binary*
    cat_df: pd.DataFrame
        DataFrame with variables that were categorized as *categorical*
    bin_df: pd.DataFrame
        DataFrame with variables that were categorized as *continuous*
    other_df: pd.DataFrame
        DataFrame with variables that were not categorized and should be examined manually

    Examples
    --------
    >>> import clarite
    >>> nhanes_bin, nhanes_cat, nhanes_cont, nhanes_other = clarite.process.categorize()
    10 of 945 variables (1.06%) had no non-NA values and are discarded.
    33 of 945 variables (3.49%) had only one value and are discarded.
    361 of 945 variables (38.20%) are classified as binary (2 values).
    44 of 945 variables (4.66%) are classified as categorical (3 to 6 values).
    461 of 945 variables (48.78%) are classified as continuous (>= 15 values).
    36 of 945 variables (3.81%) are not classified (between 6 and 15 values).
    """
    # Validate parameters
    assert cat_min > 2
    assert cat_min <= cat_max
    assert cont_min > cat_max

    # Create filter series
    num_before = len(data.columns)
    unique_count = data.nunique()

    # No values (All NA)
    zero_filter = unique_count == 0
    num_zero = sum(zero_filter)
    print(f"{num_zero:,} of {num_before:,} variables ({num_zero/num_before:.2%}) had no non-NA values and are discarded.")

    # Single value variables (useless for regression)
    single_filter = unique_count == 1
    num_single = sum(single_filter)
    print(f"{num_single:,} of {num_before:,} variables ({num_single/num_before:.2%}) had only one value and are discarded.")

    # Binary
    binary_filter = unique_count == 2
    num_binary = sum(binary_filter)
    print(f"{num_binary:,} of {num_before:,} variables ({num_binary/num_before:.2%}) are classified as binary (2 values).")
    bin_df = data.loc[:, binary_filter]

    # Categorical
    cat_filter = (unique_count >= cat_min) & (unique_count <= cat_max)
    num_cat = sum(cat_filter)
    print(f"{num_cat:,} of {num_before:,} variables ({num_cat/num_before:.2%}) are classified as categorical ({cat_min} to {cat_max} values).")
    cat_df = data.loc[:, cat_filter]

    # Continuous
    cont_filter = unique_count >= cont_min
    num_cont = sum(cont_filter)
    print(f"{num_cont:,} of {num_before:,} variables ({num_cont/num_before:.2%}) are classified as continuous (>= {cont_min} values).")
    cont_df = data.loc[:, cont_filter]

    # Other
    other_filter = ~zero_filter & ~single_filter & ~binary_filter & ~cat_filter & ~cont_filter
    num_other = sum(other_filter)
    print(f"{num_other:,} of {num_before:,} variables ({num_other/num_before:.2%}) are not classified (between {cat_max} and {cont_min} values).")
    other_df = data.loc[:, other_filter]

    return bin_df, cat_df, cont_df, other_df


def merge_variables(left: pd.DataFrame, right: pd.DataFrame, how: str = 'outer'):
    """
    Merge a list of dataframes with different variables side-by-side.  Keep all observations ('outer' merge) by default.

    Parameters
    ----------
    left: pd.Dataframe
        "left" DataFrame
    right: pd.DataFrame
        "right" DataFrame which uses the same index
    how: merge method, one of {'left', 'right', 'inner', 'outer'}
        Keep only rows present in the left data, the right data, both datasets, or either dataset.

    Examples
    --------
    >>> import clarite
    >>> df = clarite.modify.merge_variables(df_bin, df_cat, how='outer')
    """
    return left.merge(right, left_index=True, right_index=True, how=how)


def move_variables(left: pd.DataFrame, right: pd.DataFrame,
                   skip: Optional[Union[str, List[str]]] = None, only: Optional[Union[str, List[str]]] = None):
    """
    Move one or more variables from one DataFrame to another

    Parameters
    ----------
    left: pd.Dataframe
        DataFrame containing the variable(s) to be moved
    right: pd.DataFrame
        DataFrame (which uses the same index) that the variable(s) will be moved to
    skip: str, list or None (default is None)
        List of variables that will *not* be moved
    only: str, list or None (default is None)
        List of variables that are the *only* ones to be moved

    Returns
    -------
    left: pd.DataFrame
        The first DataFrame with the variables removed
    right: pd.DataFrame
        The second DataFrame with the variables added

    Examples
    --------
    >>> import clarite
    >>> df_cat, df_cont = clarity.process.move_variables(df_cat, df_cont, only=["DRD350AQ", "DRD350DQ", "DRD350GQ"])
    Moved 3 variables.
    >>> discovery_check, discovery_cont = clarite.process.move_variables(discovery_check, discovery_cont)
    Moved 39 variables.
    """
    # Which columns
    columns = _validate_skip_only(list(left), skip, only)

    # Add to new df
    right = merge_variables(right, left[columns])

    # Remove from original
    left = left.drop(columns, axis='columns')

    # Log
    if len(columns) == 1:
        print("Moved 1 variable.")
    else:
        print(f"Moved {len(columns)} variables.")

    # Return
    return left, right



from typing import Optional, Union, Dict
import numpy as np
import pandas as pd


class SurveyDesignSpec:
    """
    Holds parameters for building a statsmodels SurveyDesign object

    Parameters
    ----------
    survey_df: pd.DataFrame
        A DataFrame containing Cluster, Strata, and/or weights data
    strata: string or None
        The name of the strata variable in the survey_df
    cluster: string or None
        The name of the cluster variable in the survey_df
    nest: bool, default False
        Whether or not the clusters are nested in the strata (The same cluster IDs are repeated in different strata)
    weights: string or dictionary(string:string)
        The name of the weights variable in the survey_df, or a dictionary mapping variable names to weight names
    single_cluster: str
        Setting controlling variance calculation in single-cluster strata
        'error': default, throw an error
        'scaled': use the average value of other strata
        'centered': use the average of all observations
        'certainty': that strata doesn't contribute to the variance

    Attributes
    ----------

    Examples
    --------
    >>> import clarite
    >>> clarite.analyze.SurveyDesignSpec(survey_df=survey_design_replication,
                                         strata="SDMVSTRA",
                                         cluster="SDMVPSU",
                                         nest=True,
                                         weights=weights_replication,
                                         single_cluster='scaled')
    """
    def __init__(self,
                 survey_df: pd.DataFrame,
                 strata: Optional[str] = None,
                 cluster: Optional[str] = None,
                 nest: bool = False,
                 weights: Union[str, Dict[str, str]] = None,
                 single_cluster: Optional[str] = 'error'):

        # Validate index
        if isinstance(survey_df.index, pd.core.index.MultiIndex):
            raise ValueError("survey_df: DataFrame must not have a multiindex")
        survey_df.index.name = "ID"

        # Store parameters
        self.survey_df = survey_df
        self.strata = strata
        self.cluster = cluster
        self.nest = nest
        self.weights = weights
        self.single_cluster = single_cluster

        self.validate_params()

    def validate_params(self):
        if self.strata is not None:
            if self.strata not in self.survey_df:
                raise KeyError(f"strata key ('{self.strata}') was not found in the survey_df")

        if self.cluster is not None:
            if self.cluster not in self.survey_df:
                raise KeyError(f"cluster key ('{self.cluster}') was not found in the survey_df")

        if type(self.weights) == "dict":
            for var_name, weight_name in self.weights.items():
                if weight_name not in self.survey_df:
                    raise KeyError(f"weights key for '{var_name}' ('{weight_name}') was not found in the survey_df")
        elif type(self.weights) == "str":
            if self.weights not in self.survey_df:
                raise KeyError(f"weights ('{weight_name}') was not found in the survey_df")

        if self.single_cluster not in {'error', 'scaled', 'certainty', 'centered'}:
            raise ValueError(f"if provided, 'single_cluster' must be one of 'error', 'scaled', 'certainty', or 'centered'.")

    def get_survey_design(self, regression_variable: Optional[str] = None, index: Optional[pd.Index] = None):
        """
        Build a survey design based on the regression variable

        Parameters
        ----------
        regression_variable : str or None
            Name of the variable being regressed.  Required if weights are variable-specific.
        index : pd.Index or None
            Data being used in the analysis whose index is a subset of the survey design indicies

        Returns
        -------
        survey_design: SurveyDesign
            SurveyDesign object used for further analysis
        index: pd.Index
            Index corresponding to all data in the design- limited by the input index or weights that are na or zero
        """
        if index is None:
            index = self.survey_df.index

        # Get weights array
        if type(self.weights) == str:
            weights = self.survey_df[self.weights]
            weights = weights.loc[index]
        elif type(self.weights) == dict:
            if regression_variable is None:
                raise ValueError("This SurveyDesignSpec uses variable-specific weights- a variable name is required to create a SurveyDesign object.")
            elif regression_variable not in self.weights:
                raise KeyError(f"The regression variable ({regression_variable}) was not found in the SurveyDesignSpec")
            else:
                weights = self.survey_df[self.weights[regression_variable]]
                weights = weights.loc[index]
        else:
            weights = None

        # Update index to remove missing, negative, or zero weights
        if weights is not None:
            weights = weights[~weights.isna() & weights > 0]
            n_removed = len(index) - len(weights)
            if n_removed > 0:
                print(f"WARNING: {regression_variable} - {n_removed} observation(s) with missing, negative, or zero weights were removed")
                index = weights.index

        # Get strata array
        if self.strata is not None:
            strata = self.survey_df[self.strata]
            strata = strata.loc[index]
        else:
            strata = None

        # Get cluster array
        if self.cluster is not None:
            cluster = self.survey_df[self.cluster]
            cluster = cluster.loc[index]
        else:
            cluster = None

        return SurveyDesign(strata=strata, cluster=cluster, weights=weights, nest=self.nest, single_cluster=self.single_cluster), index


class SurveyDesign(object):
    """
    Description of a survey design, used by most methods
    implemented in this module.

    Parameters
    -------
    strata : array-like or None
        Strata for each observation. If none, an array
        of ones is constructed
    cluster : array-like or None
        Cluster for each observation. If none, an array
        of ones is constructed
    weights : array-like or None
        The weight for each observation. If none, an array
        of ones is constructed
    fpc : ps.Series or None
        The finite population correction for each observation.
        If none, an array of zeros is constructed.
        If < 1, treated as sampling weights and use directly
        Otherwise, treat as cluster population size and convert to sampling weights as (number of obs in cluster) / fpc
    nest : boolean
        allows user to specify if PSU's with the same
        PSU number in different strata are treated as distinct PSUs.
    single_cluster : str
        Setting controlling variance calculation in single-cluster strata
        'error': default, throw an error
        'scaled': use the average value of other strata
        'centered': use the average of all observations
        'certainty': that strata doesn't contribute to the variance

    Attributes
    ----------
    weights : (n, ) pd.Series
        The weight for each observation
    n_strat : integer
        The number of district strata
    clust : (n, ) pd.Series
        The relabeled cluster array from 0, 1, ..
    strat : (n, ) pd.Series
        The related strata array from 0, 1, ...
    clust_per_strat : (self.n_strat, ) array
        Holds the number of clusters in each stratum
    strat_for_clust : ndarray
        The stratum for each cluster
    n_strat: integer
        The total number of strata
    n_clust : integer
        The total number of clusters across strata
    strat_names : list[str]
        The names of strata
    clust_names : list[str]
        The names of clusters
    """

    def __init__(self,
                 strata: Optional[pd.Series] = None,
                 cluster: Optional[pd.Series] = None,
                 weights: Optional[pd.Series] = None,
                 fpc: Optional[pd.Series] = None,
                 nest: bool = True,
                 single_cluster: str = 'error'):

        # Record inputs
        # Note: Currently allowed combinations of parameters:
        #        * Strata, Clusters (weights assumed to be 1)
        #        * Strata, Clustes, Weight
        # Note: Not Currently allowed combinations of parameters:
        #        * FPC (not allowed to be provided- further testing and documentation needed)
        #        * Weight only (further testing needed to confirm correctness)
        #        * Strata or Clusters (not both) - requires further testing
        self.has_strata = False
        self.has_clusters = False
        self.has_weight = False
        self.has_fpc = False
        if strata is not None:
            self.has_strata = True
        if cluster is not None:
            self.has_clusters = True
        if weights is not None:
            self.has_weights = True
        if fpc is not None:
            self.has_fpc = True
            raise NotImplementedError(f"Providing fpc is not yet supported")
        if not self.has_strata or not self.has_clusters:
            raise NotImplementedError(f"In this version of clarite, both strata and clusters must be provided.")

        strata, cluster, self.weights, self.fpc = self._check_args(strata, cluster, weights, fpc)

        # If requested, recode the PSUs to be sure that the same PSU # in
        # different strata are treated as distinct PSUs. This is the same
        # as the nest option in R.
        if nest:
            cluster = strata.astype(str) + "-" + cluster.astype(str)

        # Make strata and cluster into categoricals
        self.strat = strata.astype('category').rename('strat')
        self.clust = cluster.astype('category').rename('clust')

        # Get a combined dataframe to map the relationships
        combined = pd.concat([self.strat, self.clust, self.fpc], axis=1)

        # The number of clusters per stratum
        self.clust_per_strat = combined.groupby('strat')['clust'].nunique()

        # The stratum for each cluster
        self.strat_for_clust = combined.groupby('clust')['strat'].unique().apply(lambda l: l[0])

        # The fpc for each cluster
        self.fpc = combined.groupby('clust')['fpc'].unique().apply(lambda l: l[0])

        # Clusters within each stratum
        self.ii = combined.groupby('strat')['clust'].unique()

        # Record names
        self.strat_names = self.strat.cat.categories
        self.clust_names = self.clust.cat.categories

        # Record number of strat/clust
        self.n_strat = len(self.strat_names)
        self.n_clust = len(self.clust_names)

        # Record single cluster setting
        self.single_cluster = single_cluster

    def __str__(self):
        """
        The __str__ method for our data
        """
        summary_list = ["Number of observations: ", str(len(self.strat)),
                        "Sum of weights: ", str(self.weights.sum()),
                        "Number of strata: ", str(self.n_strat),
                        "Clusters per stratum: ", str(self.clust_per_strat)]

        return "\n".join(summary_list)

    def _check_args(self,
                    strata: Optional[pd.Series] = None,
                    cluster: Optional[pd.Series] = None,
                    weights: Optional[pd.Series] = None,
                    fpc: Optional[pd.Series] = None):
        """
        Minor error checking to make sure user supplied any of
        strata, cluster, or weights. For unspecified subgroup labels
        an array of ones is created

        Parameters
        ----------
        strata : pd.Series or None
            Strata for each observation. If none, an array
            of ones is constructed
        cluster : pd.Series or None
            Cluster for each observation. If none, an array
            of ones is constructed
        weights : pd.Series or None
            The weight for each observation. If none, an array
            of ones is constructed
        fpc: pd.Series or None
            The finite population correction for each observation.
            If none, an array of zeros is constructed.
            If < 1, treated as sampling weights and use directly
            Otherwise, treat as cluster population size and convert to sampling weights as (number of obs in cluster) / fpc

        Returns
        -------
        strata : ndarray
            Series of the strata labels
        cluster : ndarray
            Series of the cluster labels
        weights : ndarray
            Series of the observation weights
        fpc: pd.Series
            Series of fpc values
        """
        # At least one must be defined
        if all([x is None for x in (strata, cluster, weights)]):
            raise ValueError("""At least one of strata, cluster, rep_weights, and weights
                             musts not be None""")

        # Get corrected index information for creating replacements for None
        index = [x.index for x in (strata, cluster, weights) if x is not None][0]
        n = len(index)

        # Make sure index is equal for all given parameters
        if not all([index.equals(x.index) for x in (strata, cluster, weights) if x is not None]):
            raise ValueError("""index of strata, cluster, and weights are not all compatible""")

        # Create default values or update name of given value
        if strata is None:
            strata = pd.Series(np.ones(n), index=index, name='strata')
        else:
            strata = strata.rename('strata')
        if cluster is None:
            cluster = pd.Series(np.ones(n), index=index, name='cluster')
        else:
            cluster = cluster.rename('cluster')
        if weights is None:
            weights = pd.Series(np.ones(n), index=index, name='weights')
        else:
            weights = weights.rename('weights')
        if fpc is None:
            fpc = pd.Series(np.zeros(n), index=index, name='fpc')
        else:
            fpc = fpc.rename('fpc')
            if not all(fpc <= 1):
                # Assume these are actual population size, and convert to a fraction
                combined = pd.merge(cluster, fpc, left_index=True, right_index=True)
                sampled_cluster_size = combined.groupby('cluster')['fpc'].transform('size')
                fpc = fpc/sampled_cluster_size
                try:
                    assert all((fpc >= 0) & (fpc <= 1))
                except AssertionError:
                    raise ValueError("Error processing FPC- either provide sampling fractions or cluster population sizes")
        return strata, cluster, weights, fpc

    def get_jackknife_rep_weights(self, dropped_clust):
        """
        Computes 'delete 1' jackknife replicate weights

        Parameters
        ----------
        dropped_clust : string
            Which cluster to leave out when computing 'delete 1' jackknife replicate weights

        Returns
        -------
        w : ndarray
            Augmented weight
        """
        # get stratum that the cluster belongs in
        s = self.strat_for_clust[dropped_clust]
        nh = self.clust_per_strat[s]
        w = self.weights.copy()
        # all weights within the stratum are modified
        w[self.strat == s] *= nh / float(nh - 1)
        # but if you're within the cluster to be removed, set as 0
        w[self.clust == dropped_clust] = 0
        return w

    def get_dof(self, X):
        """
        Calculate degrees of freedom based on a subset of the design

        Parameters
        ----------
        X : pd.DataFrame
            Input data used in the calculation

        Returns
        -------
        dof : int
            Degrees of freedom
        """
        # num of clusters minus num of strata minus (num of predictors - 1)
        return self.clust.loc[X.index].nunique() - self.strat.loc[X.index].nunique() - (X.shape[1] - 1)



"""
Analyze
========

EWAS and associated calculations

  .. autosummary::
     :toctree: modules/analyze

     ewas
     add_corrected_pvalues


"""

from typing import List, Optional

import pandas as pd
from numpy import nan
from statsmodels.stats.multitest import multipletests
from .survey import SurveyDesignSpec

from .modify import make_binary, make_categorical, make_continuous
from ..internal.regression import Regression


result_columns = ['variable_type', 'converged', 'N', 'beta', 'SE', 'var_pvalue', 'LRT_pvalue', 'diff_AIC', 'pvalue']
corrected_pvalue_columns = ['pvalue_bonferroni', 'pvalue_fdr']


def ewas(
        phenotype: str,
        covariates: List[str],
        bin_df: Optional[pd.DataFrame],
        cat_df: Optional[pd.DataFrame],
        cont_df: Optional[pd.DataFrame],
        survey_design_spec: Optional[SurveyDesignSpec] = None,
        cov_method: Optional[str] = 'stata',
        min_n: Optional[int] = 200):
    """
    Run an EWAS on a phenotype

    Parameters
    ----------
    phenotype: string
        The variable to be used as the output of the regressions
    covariates: list (strings),
        The variables to be used as covariates.  Any variables in the DataFrames not listed as covariates are regressed.
    bin_df: pd.DataFrame or None
        A DataFrame containing binary variables
    cat_df: pd.DataFrame or None
        A DataFrame containing categorical variables
    cont_df: pd.DataFrame or None
        A DataFrame containing continuous variables
    survey_design_spec: SurveyDesignSpec or None
        A SurveyDesignSpec object is used to create SurveyDesign objects for each regression.
    cov_method: str or None
        Covariance calculation method (if survey_design_spec is passed in).  'stata' or 'jackknife'
    min_n: int or None
        Minimum number of complete-case observations (no NA values for phenotype, covariates, variable, or weight)
        Defaults to 200

    Returns
    -------
    df: pd.DataFrame
        EWAS results DataFrame with these columns: ['variable_type', 'N', 'beta', 'SE', 'var_pvalue', 'LRT_pvalue', 'diff_AIC', 'pvalue']

    Examples
    --------
    >>> ewas_discovery = clarite.analyze.ewas("logBMI", covariates, nhanes_discovery_bin, nhanes_discovery_cat, nhanes_discovery_cont)
    Running EWAS on a continuous variable
    """
    # Make sure the index of each dataset is not a multiindex and give it a consistent name
    for param_name, df in {'bin_df': bin_df, 'cat_df': cat_df, 'cont_df': cont_df}.items():
        if df is not None:
            if isinstance(df.index, pd.core.index.MultiIndex):
                raise ValueError(f"{param_name}: DataFrame must not have a multiindex")
            df.index.name = "ID"

    # Collects lists of regression variables and format dataframes using the new index from 'unified'
    rv_bin, rv_cat, rv_cont = list(), list(), list()
    if bin_df is not None:
        # Get original columns with new index and format as binary
        bin_df = make_binary(bin_df)
        rv_bin = [v for v in list(bin_df) if v not in covariates and v != phenotype]
    if cat_df is not None:
        # Get original columns with new index and format as categorical
        cat_df = make_categorical(cat_df)
        rv_cat = [v for v in list(cat_df) if v not in covariates and v != phenotype]
    if cont_df is not None:
        # Get original columns with new index and format as continuous
        cont_df = make_continuous(cont_df)
        rv_cont = [v for v in list(cont_df) if v not in covariates and v != phenotype]

    # Ensure covariates are all present
    variables = set(list(bin_df) + list(cat_df) + list(cont_df))
    missing_covariates = [c for c in covariates if c not in variables]
    if len(missing_covariates) > 0:
        raise ValueError(f"One or more covariates were not found in any of the input DataFrames: {', '.join(missing_covariates)}")

    # Ensure there are variables which can be regressed
    if len(rv_bin + rv_cat + rv_cont) == 0:
        raise ValueError(f"No variables are available to run regression on")

    # Figure out kind of phenotype
    if phenotype in list(bin_df):
        pheno_kind = 'binary'
    elif phenotype in list(cat_df):
        pheno_kind = 'categorical'
        raise NotImplementedError("Categorical Phenotypes are not yet supported")
    elif phenotype in list(cont_df):
        pheno_kind = 'continuous'
    else:
        raise ValueError(f"Couldn't find the phenotype ('{phenotype}') in the data.")
    print(f"Running EWAS on a {pheno_kind} variable")

    # Merge dfs if there are multiple
    dfs = [df for df in [bin_df, cat_df, cont_df] if df is not None]
    if len(dfs) == 1:
        df = dfs[0]
    else:
        df = dfs[0].join(dfs[1:], how="outer")

    # Run Regressions
    return run_regressions(phenotype, covariates, df, rv_bin, rv_cat, rv_cont, pheno_kind, min_n, survey_design_spec, cov_method)


def run_regressions(phenotype: str,
                    covariates: List[str],
                    df: pd.DataFrame,
                    rv_bin: List[str],
                    rv_cat: List[str],
                    rv_cont: List[str],
                    pheno_kind: str,
                    min_n: int,
                    survey_design_spec: Optional[SurveyDesignSpec],
                    cov_method: Optional[str]):
    """Run a regressions on variables"""
    result = []

    # Must ensure phenotype is numerical for logistic regression
    if pheno_kind == "binary" or pheno_kind == "categorical":
        df[phenotype] = df[phenotype].cat.codes

    # Continuous Variables
    print(f"\n####### Regressing {len(rv_cont)} Continuous Variables #######\n")
    for rv in rv_cont:
        # Set up the regression
        regression = Regression(variable=rv,
                                variable_kind='continuous',
                                phenotype=phenotype,
                                phenotype_kind=pheno_kind,
                                data=df,
                                covariates=covariates,
                                survey_design_spec=survey_design_spec,
                                cov_method=cov_method)
        # Run the regression
        try:
            regression.run(min_n=min_n)
        except Exception as e:
            print(f"{rv} = NULL due to: {e}")
        # Save results
        result.append(regression.get_results())

    print(f"\n####### Regressing {len(rv_bin)} Binary Variables #######\n")
    for rv in rv_bin:
        # Set up the regression
        regression = Regression(variable=rv,
                                variable_kind='binary',
                                phenotype=phenotype,
                                phenotype_kind=pheno_kind,
                                data=df,
                                covariates=covariates,
                                survey_design_spec=survey_design_spec,
                                cov_method=cov_method)
        # Run the regression
        try:
            regression.run(min_n=min_n)
        except Exception as e:
            print(f"{rv} = NULL due to: {e}")
        # Save results
        result.append(regression.get_results())

    print(f"\n####### Regressing {len(rv_cat)} Categorical Variables #######\n")
    for rv in rv_cat:
        # Set up the regression
        regression = Regression(variable=rv,
                                variable_kind='categorical',
                                phenotype=phenotype,
                                phenotype_kind=pheno_kind,
                                data=df,
                                covariates=covariates,
                                survey_design_spec=survey_design_spec,
                                cov_method=cov_method)
        # Run the regression
        try:
            regression.run(min_n=min_n)
        except Exception as e:
            print(f"{rv} = NULL due to: {e}")
        # Save results
        result.append(regression.get_results())

    # Gather All Results
    result = pd.DataFrame(result)
    result['phenotype'] = phenotype  # Add phenotype
    result = result.sort_values('pvalue').set_index(['variable', 'phenotype'])  # Sort and set index
    result = result[['variable_type', 'converged', 'N', 'beta', 'SE', 'var_pvalue', 'LRT_pvalue', 'diff_AIC', 'pvalue']]  # Sort columns
    print("Completed EWAS\n")
    return result


def add_corrected_pvalues(ewas_result):
    """
    Add bonferroni and FDR pvalues to an ewas result and sort by increasing FDR (in-place)

    Parameters
    ----------
    ewas_result: pd.DataFrame
        EWAS results DataFrame with these columns: ['variable_type', 'converged', N', 'beta', 'SE', 'var_pvalue', 'LRT_pvalue', 'diff_AIC', 'pvalue']

    Returns
    -------
    None

    Examples
    --------
    >>> clarite.analyze.add_corrected_pvalues(ewas_discovery)
    """
    # NA by default
    ewas_result['pvalue_bonferroni'] = nan
    ewas_result['pvalue_fdr'] = nan
    if (~ewas_result['pvalue'].isna()).sum() > 0:
        # Calculate values, ignoring NA pvalues
        ewas_result.loc[~ewas_result['pvalue'].isna(), 'pvalue_bonferroni'] = multipletests(ewas_result.loc[~ewas_result['pvalue'].isna(), 'pvalue'],
                                                                                            method="bonferroni")[1]
        ewas_result.loc[~ewas_result['pvalue'].isna(), 'pvalue_fdr'] = multipletests(ewas_result.loc[~ewas_result['pvalue'].isna(), 'pvalue'],
                                                                                     method="fdr_bh")[1]
        ewas_result.sort_values(by=['pvalue_fdr', 'pvalue', 'converged'], inplace=True)



# flake8: noqa
"""
Survey
========

Complex survey design

  .. autosummary::
     :toctree: modules/survey

     SurveyDesignSpec

"""

from .survey_design import SurveyDesignSpec
from .survey_model import SurveyModel



import click
from ...modules import modify, io
from ...internal.utilities import _validate_skip_only
from ..parameters import input_file, output_file, skip, only


@click.group(name='modify')
def modify_cli():
    pass


@modify_cli.command(help="Remove some columns from a dataset")
@click.argument('data', type=input_file)
@click.argument('output', type=output_file)
@skip
@only
def colfilter(data, output, skip, only):
    """Load Data, remove some columns, and save the data"""
    # Load Data
    data = io.load_data(filename=data)
    # Process skip/only parameters
    columns = _validate_skip_only(list(data), skip, only)
    data = data[columns]
    # Save
    io.save(data, filename=output)
    # Log
    click.echo(click.style(f"Done: Saved {len(data.columns):,} variables with {len(data):,} observations to {output}", fg='green'))


@modify_cli.command(help="Filter variables based on the fraction of observations with a value of zero")
@click.argument('data', type=input_file)
@click.argument('output', type=output_file)
@click.option('-p', '--filter-percent', default='90.0', type=click.FloatRange(min=0, max=100),
              help="Remove variables when the percentage of observations equal to 0 is >= this value (0 to 100)")
@skip
@only
def colfilter_percent_zero(data, output, filter_percent, skip, only):
    # Load data
    data = io.load_data(data)
    # Modify
    result = modify.colfilter_percent_zero(data=data, filter_percent=filter_percent, skip=skip, only=only)
    # Save
    io.save(result, filename=output)
    # Log
    click.echo(click.style(f"Done: Saved filtered data to {output}", fg='green'))


@modify_cli.command(help="Filter variables based on a minimum number of non-NA observations")
@click.argument('data', type=input_file)
@click.argument('output', type=output_file)
@click.option('-n', default=200, type=click.IntRange(min=0),
              help="Remove variables with less than this many non-na observations")
@skip
@only
def colfilter_min_n(data, output, n, skip, only):
    # Load data
    data = io.load_data(data)
    # Modify
    result = modify.colfilter_min_n(data=data, n=n, skip=skip, only=only)
    # Save
    io.save(result, filename=output)
    # Log
    click.echo(click.style(f"Done: Saved filtered data to {output}", fg='green'))


@modify_cli.command(help="Filter variables based on a minimum number of non-NA observations per category")
@click.argument('data', type=input_file)
@click.argument('output', type=output_file)
@click.option('-n', default=200, type=click.IntRange(min=0),
              help="Remove variables with less than this many non-na observations in each category")
@skip
@only
def colfilter_min_cat_n(data, output, n, skip, only):
    # Load data
    data = io.load_data(data)
    # Modify
    result = modify.colfilter_min_cat_n(data=data, n=n, skip=skip, only=only)
    # Save
    io.save(result, filename=output)
    # Log
    click.echo(click.style(f"Done: Saved filtered data to {output}", fg='green'))


@modify_cli.command(help="Select some rows from a dataset using a simple comparison, keeping rows where the comparison is True.")
@click.argument('data', type=input_file)
@click.argument('output', type=output_file)
@click.argument('column', type=click.STRING)
@click.option('--value-str', 'vs', type=click.STRING, default=None, help="Compare values in the column to this string")
@click.option('--value-int', 'vi', type=click.INT, default=None, help="Compare values in the column to this integer")
@click.option('--value-float', 'vf', type=click.FLOAT, default=None, help="Compare values in the column to this floating point number")
@click.option('--comparison', '-c', default='eq', type=click.Choice(['lt', 'lte', 'eq', 'gte', 'gt']),
              help="Keep rows where the value of the column is lt (<), lte (<=), eq (==), gte (>=), or gt (>) the specified value.  Eq by default.")
def rowfilter(data, output, column, vs, vi, vf, comparison):
    """Load Data, keep certain rows, and save the data"""
    # Load Data
    data = io.load_data(filename=data)
    # Ensure column is present
    if column not in data.columns:
        raise ValueError(f"The specified column {column} was not found in the data.")
    # Decode value
    values = [v for v in (vs, vi, vf) if v is not None]
    if len(values) != 1:
        raise ValueError("The comparison value ('--value-str', '--value-int', or '--value-float') must be specified just once.")
    else:
        value = values[0]
    # Filter
    if comparison == 'lt':
        result = data.loc[data[column] < value, ]
    elif comparison == 'lte':
        result = data.loc[data[column] <= value, ]
    elif comparison == 'eq':
        result = data.loc[data[column] == value, ]
    elif comparison == 'gt':
        result = data.loc[data[column] >= value, ]
    elif comparison == 'gte':
        result = data.loc[data[column] > value, ]
    # Save
    io.save(result, filename=output)
    # Log
    click.echo(click.style(f"Done: Saved {len(result.columns):,} variables with {len(result):,} observations to {output}", fg='green'))


@modify_cli.command(help="Filter out observations that are not complete cases (contain no NA values)")
@click.argument('data', type=input_file)
@click.argument('output', type=output_file)
@skip
@only
def rowfilter_incomplete_obs(data, output, skip, only):
    # Load data
    data = io.load_data(data)
    # Modify
    result = modify.rowfilter_incomplete_obs(data=data, skip=skip, only=only)
    # Save
    io.save(result, filename=output)
    # Log
    click.echo(click.style(f"Done: Saved filtered data to {output}", fg='green'))


@modify_cli.command(help="Replace values in the data with other values."
                         "The value being replaced ('current') and the new value ('replacement') "
                         "are specified with their type, and only one may be included for each. "
                         "If it is not specified, the value being replaced or being inserted is None.")
@click.argument('data', type=input_file)
@click.argument('output', type=output_file)
@click.option('--current-str', 'cs', type=click.STRING, default=None, help="Replace occurences of this string value")
@click.option('--current-int', 'ci', type=click.INT, default=None, help="Replace occurences of this integer value")
@click.option('--current-float', 'cf', type=click.FLOAT, default=None, help="Replace occurences of this float value")
@click.option('--replacement-str', 'rs', type=click.STRING, default=None, help="Insert this string value")
@click.option('--replacement-int', 'ri', type=click.INT, default=None, help="Insert this integer value")
@click.option('--replacement-float', 'rf', type=click.FLOAT, default=None, help="Insert this float value")
@skip
@only
def recode_values(data, output, cs, ci, cf, rs, ri, rf, skip, only):
    # Load data
    data = io.load_data(data)
    # Decode current
    c = [v for v in (cs, ci, cf) if v is not None]
    if len(c) == 0:
        current = None
    elif len(c) == 1:
        current = c[0]
    else:
        raise ValueError("The 'current' value was specified for multiple types.  It should be specified with at most one type.")
    # Decode replacement
    r = [v for v in (rs, ri, rf) if v is not None]
    if len(r) == 0:
        replacement = None
    elif len(r) == 1:
        replacement = r[0]
    else:
        raise ValueError("The 'replacement' value was specified for multiple types.  It should be specified with at most one type.")
    # Modify
    result = modify.recode_values(data=data, replacement_dict={current: replacement}, skip=skip, only=only)
    # Save
    io.save(result, filename=output)
    # Log
    click.echo(click.style(f"Done: Saved recoded data to {output}", fg='green'))


@modify_cli.command(help="Replace outlier values with NaN.  Outliers are defined using a gaussian or IQR approach.")
@click.argument('data', type=input_file)
@click.argument('output', type=output_file)
@click.option('--method', '-m', type=click.Choice(['gaussian', 'iqr']), default='gaussian')
@click.option('--cutoff', '-c', type=click.FLOAT, default=3.0)
@skip
@only
def remove_outliers(data, output, method, cutoff, skip, only):
    # Load data
    data = io.load_data(data)
    # Modify
    result = modify.remove_outliers(data=data, method=method, cutoff=cutoff, skip=skip, only=only)
    # Save
    io.save(result, filename=output)
    # Log
    click.echo(click.style(f"Done: Removed outliers and saved data to {output}", fg='green'))


@modify_cli.command(help="Set the type of variables to 'binary'")
@click.argument('data', type=input_file)
@click.argument('output', type=output_file)
@skip
@only
def make_binary(data, output, skip, only):
    # Load data
    data = io.load_data(data)
    # Modify
    result = modify.make_binary(data=data, skip=skip, only=only)
    # Save
    io.save(result, filename=output)
    # Log
    click.echo(click.style(f"Done: Saved filtered data to {output}", fg='green'))


@modify_cli.command(help="Set the type of variables to 'categorical'")
@click.argument('data', type=input_file)
@click.argument('output', type=output_file)
@skip
@only
def make_categorical(data, output, skip, only):
    # Load data
    data = io.load_data(data)
    # Modify
    result = modify.make_categorical(data=data, skip=skip, only=only)
    # Save
    io.save(result, filename=output)
    # Log
    click.echo(click.style(f"Done: Saved filtered data to {output}", fg='green'))


@modify_cli.command(help="Set the type of variables to 'continuous'")
@click.argument('data', type=input_file)
@click.argument('output', type=output_file)
@skip
@only
def make_continuous(data, output, skip, only):
    # Load data
    data = io.load_data(data)
    # Modify
    result = modify.make_continuous(data=data, skip=skip, only=only)
    # Save
    io.save(result, filename=output)
    # Log
    click.echo(click.style(f"Done: Saved filtered data to {output}", fg='green'))


@modify_cli.command(help="Apply a function to each value of a variable")
@click.argument('data', type=input_file)
@click.argument('output', type=output_file)
@click.argument('variable', type=click.STRING)
@click.argument('transform', type=click.STRING)
@click.argument('new_name', type=click.STRING)
def transform_variable(data, output, variable, transform, new_name):
    # Load data
    data = io.load_data(data)
    # Create new variable
    try:
        data[new_name] = data[variable].apply(transform)
    except Exception:
        raise ValueError(f"Couldn't apply a function named '{transform}'' to '{variable}'' in order to create a new '{new_name}' variable.")
    # Drop old variable
    if new_name != variable:
        data = data.drop(variable, axis='columns')
    # Save
    io.save(data, filename=output)
    # Log
    click.echo(click.style(f"Done: Saved modified data to {output}", fg='green'))



from pathlib import Path

import click
import pandas as pd
from matplotlib import pyplot as plt

from ...modules import plot, io
from ...modules.analyze import result_columns, corrected_pvalue_columns
from ..parameters import input_file, output_file


@click.group(name='plot')
def plot_cli():
    pass


@plot_cli.command(help="Create a histogram plot of a variable")
@click.argument('data', type=input_file)
@click.argument('output', type=output_file)
@click.argument('variable', type=click.STRING)
def histogram(data, output, variable):
    # Load data
    data = io.load_data(data)
    # Plot
    plot.histogram(data=data, column=variable)
    # Save and Close
    plt.savefig(output)
    plt.close()
    # Log
    click.echo(click.style(f"Done: Saved plot to {output}", fg='green'))


@plot_cli.command(help="Generate a pdf containing distribution plots for each variable")
@click.argument('data', type=input_file)
@click.argument('output', type=output_file)
@click.option('--kind', '-k', default='count', type=click.Choice(['count', 'box', 'violin', 'qq']),
              help="Kind of plot used for continuous data.  Non-continuous always shows a count plot.")
@click.option('--nrows', default=4, type=click.IntRange(min=1, max=10), help="Number of rows per page")
@click.option('--ncols', default=3, type=click.IntRange(min=1, max=10), help="Number of columns per page")
@click.option('--quality', '-q', default='medium', type=click.Choice(['low', 'medium', 'high']),
              help="Quality of the generated plots: low (150 dpi), medium (300 dpi), or high (1200 dpi).")
@click.option('--sort/--no-sort', help="Sort variables alphabetically")
def distributions(data, output, kind, nrows, ncols, quality, sort):
    # Load data
    data = io.load_data(data)
    # Plot and save
    plot.distributions(data=data, filename=output, continuous_kind=kind, nrows=nrows, ncols=ncols, quality=quality, sort=sort)
    # Log
    click.echo(click.style(f"Done: Saved plot to {output}", fg='green'))


@plot_cli.command(help="Generate a manhattan plot of EWAS results")
@click.argument('data', type=input_file)
@click.argument('output', type=output_file)
@click.option('--categories', '-c', type=input_file, default=None, help="tab-separate file with two columns: 'Variable' and 'category'")
@click.option('--other', '-o', multiple=True, help="other datasets to include in the plot")
@click.option('--nlabeled', default=3, type=click.IntRange(min=0, max=50), help="label top n points")
@click.option('--label', default=None, multiple=True, type=click.STRING, help="label points by name")
def manhattan(data, output, categories, other, nlabeled, label):
    # Load data
    data = {Path(data).name: pd.read_csv(data, sep="\t", index_col=['variable', 'phenotype'])}
    for o in other:
        data[Path(o).name] = pd.read_csv(o, sep="\t", index_col=['variable', 'phenotype'])
    for d_name, d in data.items():
        if list(d) != result_columns + corrected_pvalue_columns:
            raise ValueError(f"{d_name} was not a valid EWAS result file.")
    # Load categories, if any
    if categories is not None:
        categories = pd.read_csv(categories, sep="\t")
        categories.columns = ['Variable', 'category']
        categories = categories.set_index('Variable')['category'].to_dict()
    # Plot and save
    plot.manhattan(data, categories=categories, num_labeled=nlabeled, label_vars=label, filename=output)
    # Log
    click.echo(click.style(f"Done: Saved plot to {output}", fg='green'))



"""
Describe
========

Functions that are used to gather information about some data

  .. autosummary::
     :toctree: modules/describe

     correlations
     freq_table
     percent_na

"""

# Describe - functions that are used to gather information about some data

import numpy as np
import pandas as pd


def correlations(data, threshold: float = 0.75):
    """
    Return variables with pearson correlation above the threshold

    Parameters
    ----------
    data: pd.DataFrame
        The DataFrame to be described
    threshold: float, between 0 and 1
        Return a dataframe listing pairs of variables whose absolute value of correlation is above this threshold

    Returns
    -------
    result: pd.DataFrame
        DataFrame listing pairs of correlated variables and their correlation value

    Examples
    --------
    >>> import clarite
    >>> correlations = clarite.describe.correlations(df, threshold=0.9)
    >>> correlations.head()
                        var1      var2  correlation
    0  supplement_count  DSDCOUNT     1.000000
    1          DR1TM181  DR1TMFAT     0.997900
    2          DR1TP182  DR1TPFAT     0.996172
    3          DRD370FQ  DRD370UQ     0.987974
    4          DR1TS160  DR1TSFAT     0.984733
    """
    # Get correlaton matrix
    correlation = data.corr()
    # Keep only the upper triangle to avoid listing both a-b and b-a correlations
    correlation = correlation.where(
        np.triu(np.ones(correlation.shape), k=1).astype(np.bool)
    )
    # Stack and rename into the desired format
    correlation = (
        correlation.stack()
        .rename("correlation")
        .rename_axis(["var1", "var2"])
        .reset_index()
    )
    # Remove those with correlation below threshold
    correlation = correlation.loc[correlation["correlation"].abs() >= threshold, ]
    # Sort by absolute value
    correlation = correlation.reindex(correlation["correlation"].abs().sort_values(ascending=False).index)
    # Return with a reset index
    return correlation.reset_index(drop=True)


def freq_table(data):
    """
    Return the count of each unique value for all categorical variables.  Non-categorical typed variables
    will return a single row with a value of '<Non-Categorical Values>' and the number of non-NA values.

    Parameters
    ----------
    data: pd.DataFrame
        The DataFrame to be described

    Returns
    -------
    result: pd.DataFrame
        DataFrame listing variable, value, and count for each categorical variable

    Examples
    --------
    >>> import clarite
    >>> clarite.describe.freq_table(df).head(n=10)
        variable value  count
    0                 SDDSRVYR                         2   4872
    1                 SDDSRVYR                         1   4191
    2                   female                         1   4724
    3                   female                         0   4339
    4  how_many_years_in_house                         5   2961
    5  how_many_years_in_house                         3   1713
    6  how_many_years_in_house                         2   1502
    7  how_many_years_in_house                         1   1451
    8  how_many_years_in_house                         4   1419
    9                  LBXPFDO  <Non-Categorical Values>   1032
    """
    # Define a function to be applied to each categorical variable
    def formatted_value_counts(var_name: str, df: pd.DataFrame):
        if str(df[var_name].dtype) == "category":
            df = (
                df[var_name]
                .value_counts()
                .reset_index()
                .rename({"index": "value", var_name: "count"}, axis="columns")
            )
            df["variable"] = var_name
            return df[["variable", "value", "count"]]  # reorder columns
        else:
            return pd.DataFrame.from_dict(
                {
                    "variable": [var_name],
                    "value": ["<Non-Categorical Values>"],
                    "count": [df[var_name].count()],
                }
            )

    return pd.concat(
        [formatted_value_counts(var_name, data) for var_name in list(data)]
    ).reset_index(drop=True)


def percent_na(data):
    """
    Return the percent of observations that are NA for each variable

    Parameters
    ----------
    data: pd.DataFrame
        The DataFrame to be described

    Returns
    -------
    result: pd.Series
        Series listing percent NA for each variable

    Examples
    --------
    >>> import clarite
    >>> clarite.describe.percent_na(df)
    variable    percent_na
    SDDSRVYR                 0.00000
    female                   0.00000
    LBXHBC                   4.99321
    LBXHBS                   4.98730
    """
    result = 100 * (1 - (data.count() / data.apply(len)))
    result = result.reset_index()
    result.columns = ['variable', 'percent_na']
    return result



import click
from ...modules import describe, io
from ..parameters import input_file, output_file


@click.group(name='describe')
def describe_cli():
    pass


@describe_cli.command(help="Report top correlations between variables")
@click.argument('data', type=input_file)
@click.argument('output', type=output_file)
@click.option('-t', '--threshold', default=0.75, help="Report correlations with R >= this value")
def correlations(data, output, threshold):
    # Load data
    data = io.load_data(data)
    # Describe
    results = describe.correlations(data, threshold)
    # Save results
    results.to_csv(output, sep="\t", index=False)
    # Log
    click.echo(click.style(f"Done: Saved {len(results):,} correlations to {output}", fg='green'))


@describe_cli.command(help="Report the number of occurences of each value for each variable")
@click.argument('data', type=input_file)
@click.argument('output', type=output_file)
def freq_table(data, output):
    # Load data
    data = io.load_data(data)
    # Describe
    results = describe.freq_table(data)
    # Save results
    results.to_csv(output, sep="\t", index=False)
    # Log
    processed = results.loc[results['value'] != '<Non-Categorical Values>', ]
    if len(processed) > 0:
        num_values = processed[['variable', 'value']].nunique()
        num_variables = processed['variable'].nunique()
    else:
        num_values = 0
        num_variables = 0
    click.echo(click.style(f"Done: Saved {num_values:,} unique value counts for {num_variables:,} non-continuous variables to {output}", fg='green'))


@describe_cli.command(help="Report the percent of observations that are NA for each variable")
@click.argument('data', type=input_file)
@click.argument('output', type=output_file)
def percent_na(data, output):
    # Load data
    data = io.load_data(data)
    # Describe
    results = describe.percent_na(data)
    # Save results
    results.to_csv(output, sep="\t", index=False)
    # Log
    click.echo(click.style(f"Done: Saved results for {len(results):,} variables to {output}", fg='green'))
